from __future__ import annotations

import cv2
import numpy as np


def crop_image(image: np.ndarray, roi: tuple[int, int, int, int]) -> np.ndarray:
    """Crop an image given a roi in proper format.

    Args:
        image: array of size HxW or HxWxC
        roi: (w_upper_left, h_upper_left, w_bottom_right, h_bottom_right)

    Returns:
        Cropped image based on roi
    """
    return image[roi[1] : roi[3], roi[0] : roi[2]]


def keep_aspect_ratio_resize(image: np.ndarray, size: int = 224, interpolation: int = 1) -> np.ndarray:
    """Resize input image while keeping its aspect ratio."""
    (h, w) = image.shape[:2]

    if h < w:
        height = size
        width = int(w * size / h)
    else:
        width = size
        height = int(h * size / w)

    resized = cv2.resize(image, (width, height), interpolation=interpolation)
    return resized



import torch


def barlowtwins_loss(
    z1: torch.Tensor,
    z2: torch.Tensor,
    lambd: float,
) -> torch.Tensor:
    """BarlowTwins loss described in https://arxiv.org/abs/2103.03230.

    Args:
        z1: First `augmented` normalized features (i.e. f(T(x))).
            The normalization can be obtained with
            z1_norm = (z1 - z1.mean(0)) / z1.std(0)
        z2: Second `augmented` normalized features (i.e. f(T(x))).
            The normalization can be obtained with
            z2_norm = (z2 - z2.mean(0)) / z2.std(0)
        lambd: lambda multiplier for redundancy term.

    Returns:
        BarlowTwins loss
    """
    z1 = (z1 - z1.mean(0)) / z1.std(0)
    z1 = (z2 - z2.mean(0)) / z2.std(0)
    cov = z1.T @ z2
    cov.div_(z1.size(0))
    n = cov.size(0)
    invariance_term = torch.diagonal(cov).add_(-1).pow_(2).sum()
    off_diag = cov.flatten()[:-1].view(n - 1, n + 1)[:, 1:].flatten()
    redundancy_term = off_diag.pow_(2).sum()
    return invariance_term + lambd * redundancy_term


class BarlowTwinsLoss(torch.nn.Module):
    """BarlowTwin loss.

    Args:
        lambd: lambda of the loss.
    """

    def __init__(self, lambd: float):
        super().__init__()
        self.lambd = lambd

    def forward(self, z1: torch.Tensor, z2: torch.Tensor) -> torch.Tensor:
        """Compute the BarlowTwins loss."""
        return barlowtwins_loss(z1, z2, self.lambd)



from __future__ import annotations

import getpass
import os
from abc import ABC, abstractmethod
from datetime import datetime
from typing import Any, Literal

from quadra.utils.utils import get_logger

log = get_logger(__name__)

try:
    import mlflow  # noqa
    from mlflow.entities import Run  # noqa
    from mlflow.entities.model_registry import ModelVersion  # noqa
    from mlflow.exceptions import RestException  # noqa
    from mlflow.tracking import MlflowClient  # noqa

    MLFLOW_AVAILABLE = True
except ImportError:
    MLFLOW_AVAILABLE = False


VERSION_MD_TEMPLATE = "## **Version {}**\n"
DESCRIPTION_MD_TEMPLATE = "### Description: \n{}\n"


class AbstractModelManager(ABC):
    """Abstract class for model managers."""

    @abstractmethod
    def register_model(
        self, model_location: str, model_name: str, description: str, tags: dict[str, Any] | None = None
    ) -> Any:
        """Register a model in the model registry."""

    @abstractmethod
    def get_latest_version(self, model_name: str) -> Any:
        """Get the latest version of a model for all the possible stages or filtered by stage."""

    @abstractmethod
    def transition_model(self, model_name: str, version: int, stage: str, description: str | None = None) -> Any:
        """Transition the model with the given version to a new stage."""

    @abstractmethod
    def delete_model(self, model_name: str, version: int, description: str | None = None) -> None:
        """Delete a model with the given version."""

    @abstractmethod
    def register_best_model(
        self,
        experiment_name: str,
        metric: str,
        model_name: str,
        description: str,
        tags: dict[str, Any] | None = None,
        mode: Literal["max", "min"] = "max",
        model_path: str = "deployment_model",
    ) -> Any:
        """Register the best model from an experiment."""

    @abstractmethod
    def download_model(self, model_name: str, version: int, output_path: str) -> None:
        """Download the model with the given version to the given output path."""


class MlflowModelManager(AbstractModelManager):
    """Model manager for Mlflow."""

    def __init__(self):
        if not MLFLOW_AVAILABLE:
            raise ImportError("Mlflow is not available, please install it with pip install mlflow")

        if os.getenv("MLFLOW_TRACKING_URI") is None:
            raise ValueError("MLFLOW_TRACKING_URI environment variable is not set")

        mlflow.set_tracking_uri(os.getenv("MLFLOW_TRACKING_URI"))
        self.client = MlflowClient()

    def register_model(
        self, model_location: str, model_name: str, description: str | None = None, tags: dict[str, Any] | None = None
    ) -> ModelVersion:
        """Register a model in the model registry.

        Args:
            model_location: The model uri
            model_name: The name of the model after it is registered
            description: A description of the model, this will be added to the model changelog
            tags: A dictionary of tags to add to the model

        Returns:
            The model version
        """
        model_version = mlflow.register_model(model_uri=model_location, name=model_name, tags=tags)
        log.info("Registered model %s with version %s", model_name, model_version.version)
        registered_model_description = self.client.get_registered_model(model_name).description

        if model_version.version == "1":
            header = "# MODEL CHANGELOG\n"
        else:
            header = ""

        new_model_description = VERSION_MD_TEMPLATE.format(model_version.version)
        new_model_description += self._get_author_and_date()
        new_model_description += self._generate_description(description)

        self.client.update_registered_model(model_name, header + registered_model_description + new_model_description)

        self.client.update_model_version(
            model_name, model_version.version, "# MODEL CHANGELOG\n" + new_model_description
        )

        return model_version

    def get_latest_version(self, model_name: str) -> ModelVersion:
        """Get the latest version of a model.

        Args:
            model_name: The name of the model

        Returns:
            The model version
        """
        latest_version = max(int(x.version) for x in self.client.get_latest_versions(model_name))
        model_version = self.client.get_model_version(model_name, latest_version)

        return model_version

    def transition_model(
        self, model_name: str, version: int, stage: str, description: str | None = None
    ) -> ModelVersion | None:
        """Transition a model to a new stage.

        Args:
            model_name: The name of the model
            version: The version of the model
            stage: The stage of the model
            description: A description of the transition, this will be added to the model changelog
        """
        previous_stage = self._safe_get_stage(model_name, version)

        if previous_stage is None:
            return None

        if previous_stage.lower() == stage.lower():
            log.warning("Model %s version %s is already in stage %s", model_name, version, stage)
            return self.client.get_model_version(model_name, version)

        log.info("Transitioning model %s version %s from %s to %s", model_name, version, previous_stage, stage)
        model_version = self.client.transition_model_version_stage(name=model_name, version=version, stage=stage)
        new_stage = model_version.current_stage
        registered_model_description = self.client.get_registered_model(model_name).description
        single_model_description = self.client.get_model_version(model_name, version).description

        new_model_description = "## **Transition:**\n"
        new_model_description += f"### Version {model_version.version} from {previous_stage} to {new_stage}\n"
        new_model_description += self._get_author_and_date()
        new_model_description += self._generate_description(description)

        self.client.update_registered_model(model_name, registered_model_description + new_model_description)
        self.client.update_model_version(
            model_name, model_version.version, single_model_description + new_model_description
        )

        return model_version

    def delete_model(self, model_name: str, version: int, description: str | None = None) -> None:
        """Delete a model.

        Args:
            model_name: The name of the model
            version: The version of the model
            description: Why the model was deleted, this will be added to the model changelog
        """
        model_stage = self._safe_get_stage(model_name, version)

        if model_stage is None:
            return

        if (
            input(
                f"Model named `{model_name}`, version {version} is in stage {model_stage}, "
                "type the model name to continue deletion:"
            )
            != model_name
        ):
            log.warning("Model name did not match, aborting deletion")
            return

        log.info("Deleting model %s version %s", model_name, version)
        self.client.delete_model_version(model_name, version)

        registered_model_description = self.client.get_registered_model(model_name).description

        new_model_description = "## **Deletion:**\n"
        new_model_description += f"### Version {version} from stage: {model_stage}\n"
        new_model_description += self._get_author_and_date()
        new_model_description += self._generate_description(description)

        self.client.update_registered_model(model_name, registered_model_description + new_model_description)

    def register_best_model(
        self,
        experiment_name: str,
        metric: str,
        model_name: str,
        description: str | None = None,
        tags: dict[str, Any] | None = None,
        mode: Literal["max", "min"] = "max",
        model_path: str = "deployment_model",
    ) -> ModelVersion | None:
        """Register the best model from an experiment.

        Args:
            experiment_name: The name of the experiment
            metric: The metric to use to determine the best model
            model_name: The name of the model after it is registered
            description: A description of the model, this will be added to the model changelog
            tags: A dictionary of tags to add to the model
            mode: The mode to use to determine the best model, either "max" or "min"
            model_path: The path to the model within the experiment run

        Returns:
            The registered model version if successful, otherwise None
        """
        if mode not in ["max", "min"]:
            raise ValueError(f"Mode must be either 'max' or 'min', got {mode}")

        experiment_id = self.client.get_experiment_by_name(experiment_name).experiment_id
        runs = self.client.search_runs(experiment_ids=[experiment_id])

        if len(runs) == 0:
            log.error("No runs found for experiment %s", experiment_name)
            return None

        best_run: Run | None = None

        # We can only make comparisons if the model is on the top folder, otherwise just check if the folder exists
        # TODO: Is there a better way to do this?
        base_model_path = model_path.split("/")[0]

        for run in runs:
            run_artifacts = [x.path for x in self.client.list_artifacts(run.info.run_id) if x.path == base_model_path]

            if len(run_artifacts) == 0:
                # If we don't find the given model path, skip this run
                continue

            if best_run is None:
                # If we find a run with the model it must also have the metric
                if run.data.metrics.get(metric) is not None:
                    best_run = run
                continue

            if mode == "max":
                if run.data.metrics[metric] > best_run.data.metrics[metric]:
                    best_run = run
            elif run.data.metrics[metric] < best_run.data.metrics[metric]:
                best_run = run

        if best_run is None:
            log.error("No runs found for experiment %s with the given metric", experiment_name)
            return None

        best_model_uri = f"runs:/{best_run.info.run_id}/{model_path}"

        model_version = self.register_model(
            model_location=best_model_uri, model_name=model_name, tags=tags, description=description
        )

        return model_version

    def download_model(self, model_name: str, version: int, output_path: str) -> None:
        """Download the model with the given version to the given output path.

        Args:
            model_name: The name of the model
            version: The version of the model
            output_path: The path to save the model to
        """
        artifact_uri = self.client.get_model_version_download_uri(model_name, version)
        log.info("Downloading model %s version %s from %s to %s", model_name, version, artifact_uri, output_path)
        if not os.path.exists(output_path):
            log.info("Creating output path %s", output_path)
            os.makedirs(output_path)
        mlflow.artifacts.download_artifacts(artifact_uri=artifact_uri, dst_path=output_path)

    @staticmethod
    def _generate_description(description: str | None = None) -> str:
        """Generate the description markdown template."""
        if description is None:
            return ""

        return DESCRIPTION_MD_TEMPLATE.format(description)

    @staticmethod
    def _get_author_and_date() -> str:
        """Get the author and date markdown template."""
        author_and_date = f"### Author: {getpass.getuser()}\n"
        author_and_date += f"### Date: {datetime.now().astimezone().strftime('%d/%m/%Y %H:%M:%S %Z')}\n"

        return author_and_date

    def _safe_get_stage(self, model_name: str, version: int) -> str | None:
        """Get the stage of a model version.

        Args:
            model_name: The name of the model
            version: The version of the model

        Returns:
            The stage of the model version if it exists, otherwise None
        """
        try:
            model_stage = self.client.get_model_version(model_name, version).current_stage
            return model_stage
        except RestException:
            log.error("Model named %s with version %s does not exist", model_name, version)
            return None



from __future__ import annotations

import glob
import os
from typing import Any, Literal

import torch
from pytorch_lightning import Callback, LightningModule, Trainer
from pytorch_lightning.callbacks import LearningRateMonitor
from pytorch_lightning.loggers import MLFlowLogger
from pytorch_lightning.utilities import rank_zero_only
from pytorch_lightning.utilities.types import STEP_OUTPUT

from quadra.utils.mlflow import get_mlflow_logger


def check_minio_credentials() -> None:
    """Check minio credentials for aws based storage such as minio.

    Returns:
        None
    """
    check = os.environ.get("AWS_ACCESS_KEY_ID") is not None and os.environ.get("AWS_SECRET_ACCESS_KEY") is not None
    if not check:
        raise ValueError(
            "You are trying to upload mlflow artifacts, but minio credentials are not set. Please set them in your"
            " environment variables."
        )


def check_file_server_dependencies() -> None:
    """Check file dependencies as boto3.

    Returns:
        None
    """
    try:
        # pylint: disable=unused-import,import-outside-toplevel
        import boto3  # noqa
        import minio  # noqa
    except ImportError as e:
        raise ImportError(
            "You are trying to upload mlflow artifacts, but boto3 and minio are not installed. Please install them by"
            " calling pip install minio boto3."
        ) from e


def validate_artifact_storage(logger: MLFlowLogger):
    """Validate artifact storage.

    Args:
        logger: Mlflow logger from pytorch lightning.

    """
    from quadra.utils.utils import get_logger  # pylint: disable=[import-outside-toplevel]

    log = get_logger(__name__)

    client = logger.experiment
    # TODO: we have to access the internal api to get the artifact uri, however there could be a better way
    artifact_uri = client._tracking_client._get_artifact_repo(  # pylint: disable=protected-access
        logger.run_id
    ).artifact_uri
    if artifact_uri.startswith("s3://"):
        check_minio_credentials()
        check_file_server_dependencies()
        log.info("Mlflow artifact storage is AWS/S3 basedand credentials and dependencies are satisfied.")
    else:
        log.info("Mlflow artifact storage uri is %s. Validation checks are not implemented.", artifact_uri)


class UploadCodeAsArtifact(Callback):
    """Callback used to upload Code as artifact.

    Uploads all *.py files to mlflow as an artifact, at the beginning of the run but
        after initializing the trainer. It creates project-source folder under mlflow
        artifacts and other necessary subfolders.

    Args:
        source_dir: Folder where all the source files are stored.
    """

    def __init__(self, source_dir: str):
        self.source_dir = source_dir

    @rank_zero_only
    def on_test_end(self, trainer: Trainer, pl_module: LightningModule):
        """Triggered at the end of test. Uploads all *.py files to mlflow as an artifact.

        Args:
            trainer: Pytorch Lightning trainer.
            pl_module: Pytorch Lightning module.
        """
        logger = get_mlflow_logger(trainer=trainer)

        if logger is None:
            return

        experiment = logger.experiment

        for path in glob.glob(os.path.join(self.source_dir, "**/*.py"), recursive=True):
            stripped_path = path.replace(self.source_dir, "")
            if len(stripped_path.split("/")) > 1:
                file_path_tree = "/" + "/".join(stripped_path.split("/")[:-1])
            else:
                file_path_tree = ""
            experiment.log_artifact(
                run_id=logger.run_id,
                local_path=path,
                artifact_path=f"project-source{file_path_tree}",
            )


class LogGradients(Callback):
    """Callback used to logs of the model at the end of the of each training step.

    Args:
        norm: Norm to use for the gradient. Default is L2 norm.
        tag: Tag to add to the gradients. If None, no tag will be added.
        sep: Separator to use in the log.
        round_to: Number of decimals to round the gradients to.
        log_all_grads: If True, log all gradients, not just the total norm.
    """

    def __init__(
        self,
        norm: int = 2,
        tag: str | None = None,
        sep: str = "/",
        round_to: int = 3,
        log_all_grads: bool = False,
    ):
        self.norm = norm
        self.tag = tag
        self.sep = sep
        self.round_to = round_to
        self.log_all_grads = log_all_grads

    def _grad_norm(self, named_params) -> dict:
        """Compute the gradient norm and return it in a dictionary."""
        grad_tag = "" if self.tag is None else "_" + self.tag
        results = {}
        for name, p in named_params:
            if p.requires_grad and p.grad is not None:
                norm = float(p.grad.data.norm(self.norm))
                key = f"grad_norm_{self.norm}{grad_tag}{self.sep}{name}"
                results[key] = round(norm, 3)
        total_norm = float(torch.tensor(list(results.values())).norm(self.norm))
        if not self.log_all_grads:
            # clear dictionary
            results = {}
        results[f"grad_norm_{self.norm}_total{grad_tag}"] = round(total_norm, self.round_to)
        return results

    @rank_zero_only
    def on_train_batch_end(
        self,
        trainer: Trainer,
        pl_module: LightningModule,
        outputs: STEP_OUTPUT,
        batch: Any,
        batch_idx: int,
        unused: int | None = 0,
    ) -> None:
        """Method called at the end of the train batch
        Args:
            trainer: pl.trainer
            pl_module: lightning module
            outputs: outputs
            batch: batch
            batch_idx: index
            unused: dl index.


        Returns:
            None
        """
        # pylint: disable=unused-argument
        logger = get_mlflow_logger(trainer=trainer)

        if logger is None:
            return

        named_params = pl_module.named_parameters()
        grads = self._grad_norm(named_params)
        logger.log_metrics(grads)


class UploadCheckpointsAsArtifact(Callback):
    """Callback used to upload checkpoints as artifacts.

    Args:
        ckpt_dir: Folder where all the checkpoints are stored in artifact folder.
        ckpt_ext: Extension of checkpoint files (default: ckpt).
        upload_best_only: Only upload best checkpoint (default: False)
        delete_after_upload: Delete the checkpoint from local storage after uploading (default: True)
        upload: If True, upload the checkpoints. If False, only save them on local machine.
    """

    def __init__(
        self,
        ckpt_dir: str = "checkpoints/",
        ckpt_ext: str = "ckpt",
        upload_best_only: bool = False,
        delete_after_upload: bool = True,
        upload: bool = True,
    ):
        self.ckpt_dir = ckpt_dir
        self.upload_best_only = upload_best_only
        self.ckpt_ext = ckpt_ext
        self.delete_after_upload = delete_after_upload
        self.upload = upload

    @rank_zero_only
    def on_test_end(self, trainer: Trainer, pl_module: LightningModule):
        """Triggered at the end of test. Uploads all model checkpoints to mlflow as an artifact.

        Args:
            trainer: Pytorch Lightning trainer.
            pl_module: Pytorch Lightning module.
        """
        logger = get_mlflow_logger(trainer=trainer)

        if logger is None:
            return

        experiment = logger.experiment

        if (
            trainer.checkpoint_callback
            and self.upload_best_only
            and hasattr(trainer.checkpoint_callback, "best_model_path")
        ):
            if self.upload:
                experiment.log_artifact(
                    run_id=logger.run_id,
                    local_path=trainer.checkpoint_callback.best_model_path,
                    artifact_path="checkpoints",
                )
        else:
            for path in glob.glob(os.path.join(self.ckpt_dir, f"**/*.{self.ckpt_ext}"), recursive=True):
                if self.upload:
                    experiment.log_artifact(
                        run_id=logger.run_id,
                        local_path=path,
                        artifact_path="checkpoints",
                    )
        if self.delete_after_upload:
            for path in glob.glob(os.path.join(self.ckpt_dir, f"**/*.{self.ckpt_ext}"), recursive=True):
                os.remove(path)


class LogLearningRate(LearningRateMonitor):
    """Learning rate logger at the end of the training step/epoch.

    Args:
        logging_interval: Logging interval.
        log_momentum: If True, log momentum as well.
    """

    def __init__(self, logging_interval: Literal["step", "epoch"] | None = None, log_momentum: bool = False):
        super().__init__(logging_interval=logging_interval, log_momentum=log_momentum)

    def on_train_batch_start(self, trainer, *args, **kwargs):
        """Log learning rate at the beginning of the training step if logging interval is set to step."""
        if not trainer.logger_connector.should_update_logs:
            return
        if self.logging_interval != "epoch":
            logger = get_mlflow_logger(trainer=trainer)

            if logger is None:
                return

            interval = "step" if self.logging_interval is None else "any"
            latest_stat = self._extract_stats(trainer, interval)

            if latest_stat:
                logger.log_metrics(latest_stat, step=trainer.global_step)

    def on_train_epoch_start(self, trainer, *args, **kwargs):
        """Log learning rate at the beginning of the epoch if logging interval is set to epoch."""
        if self.logging_interval != "step":
            interval = "epoch" if self.logging_interval is None else "any"
            latest_stat = self._extract_stats(trainer, interval)
            logger = get_mlflow_logger(trainer=trainer)

            if logger is None:
                return

            if latest_stat:
                logger.log_metrics(latest_stat, step=trainer.global_step)



from __future__ import annotations

import math

import torch

from quadra.schedulers.base import LearningRateScheduler
from quadra.utils.utils import get_logger

log = get_logger(__name__)


def cosine_annealing_with_warmup(
    init_lrs: list[float],
    step: int,
    total_steps: int,
    warmup_steps: int,
    lr_reduce_factor: float = 0.001,
) -> list[float]:
    """Cosine learning rate scheduler with linear warmup helper function.

    Args:
        init_lrs: The initial learning rate, one for every `param_group`.
        step: the current step
        total_steps: the total steps
        warmup_steps: total linear warmup steps
        lr_reduce_factor: reduce factor for the initial learning
            rate. This is used to set the minimum learning rate as
            `init_lr[i] * lr_reduce_factor`
            Defaults to 0.001.

    Returns:
        Annealed learning rate for this `step`
    """
    lrs = []
    for init_lr in init_lrs:
        if step < warmup_steps:
            lr = init_lr * step / warmup_steps
        else:
            step -= warmup_steps
            total_steps -= warmup_steps
            q = 0.5 * (1 + math.cos(math.pi * step / total_steps))
            end_lr = init_lr * lr_reduce_factor
            lr = init_lr * q + end_lr * (1 - q)
        lrs.append(lr)
    return lrs


class CosineAnnealingWithLinearWarmUp(LearningRateScheduler):
    """Cosine learning rate scheduler with linear warmup.

    Args:
        optimizer: optimizer for which the learning rate
            has to be optimized. If your are using this scheduler, than you have
            set the learning rate of the optimizer to 0
        batch_size: global batch size of the data loader. For more information please take a look at
            https://pytorch-lightning.readthedocs.io/en/latest/advanced/multi_gpu.html?highlight=batch%20size#batch-size
        total_epochs: the total number of epochs
        init_lr: The initial learning rate, one for every `param_group`. Mind that the
            learning rate it's linearly scaled by `batch_size` / `lr_scale`, as
            specified by https://arxiv.org/abs/1706.02677.
            Defaults to 0.01.
        lr_scale: the learning rate scheduler. Mind that the learning rate it's linearly
            scaled by `batch_size` / `lr_scale` as specified by https://arxiv.org/abs/1706.02677.
            Defaults to 256.
        linear_warmup_epochs: how many epochs for the initial
            linear learning rate scaling.
            Defaults to 10.
        lr_reduce_factor: factor to be multiplied by scaled
            lr (init_lr * batch_size / lr_scale) to avoid reaching 0 lr
            at the end of training.
        len_loader: number of batches in a given dataloader. Remind that
            the `len_loader` must be divided by total number of gpus used during the
            training. If one specifies the `len_loader` parameter, then the unit measure
            for the lr update will be in steps (number of batches), not in epochs.
            Defaults to None.
        scheduler_interval: 'step' or 'epoch'. If 'step' then the
            scheduler expects 'len_loader' to be not None.
            Defaults to `epoch`.
    """

    def __init__(
        self,
        optimizer: torch.optim.Optimizer,
        batch_size: int,
        total_epochs: int,
        init_lr: tuple[float, ...] = (0.01,),
        lr_scale: float = 256.0,
        linear_warmup_epochs: int = 10,
        lr_reduce_factor: float = 0.001,
        len_loader: int | None = None,
        scheduler_interval: str = "epoch",
    ) -> None:
        super().__init__(optimizer, init_lr)
        assert batch_size > 0
        assert total_epochs > 0
        assert lr_scale != 0
        assert linear_warmup_epochs >= 0
        assert lr_reduce_factor > 0
        assert scheduler_interval.lower() in ["step", "epoch"]
        self.batch_size = batch_size
        self.total_epochs = total_epochs
        self.linear_warmup_epochs = linear_warmup_epochs
        self.base_lr_scale = self.batch_size / lr_scale
        self.updates_counter = 1
        self.init_lr = tuple(lr * self.base_lr_scale for lr in init_lr)
        self.lr = self.init_lr
        self.lr_reduce_factor = lr_reduce_factor

        self.scheduler_interval = scheduler_interval.lower()
        if self.scheduler_interval == "step":
            assert len_loader is not None and len_loader > 0
            self.total_epochs = len_loader * self.total_epochs
            self.linear_warmup_epochs = len_loader * self.linear_warmup_epochs

    def step(self):
        """Update the learning rate for the current step."""
        self.lr = cosine_annealing_with_warmup(
            self.init_lr,
            self.updates_counter,
            self.total_epochs,
            self.linear_warmup_epochs,
            self.lr_reduce_factor,
        )
        self.set_lr(self.lr)
        self.updates_counter += 1
        return self.lr



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, cast

import numpy as np
import torch
from hydra.utils import instantiate
from omegaconf import DictConfig, OmegaConf
from torch import nn
from torch.jit import RecursiveScriptModule

from quadra.utils.logger import get_logger

try:
    import onnxruntime as ort  # noqa

    ONNX_AVAILABLE = True
except ImportError:
    ONNX_AVAILABLE = False


log = get_logger(__name__)


class BaseEvaluationModel(ABC):
    """Base interface for all evaluation models."""

    def __init__(self, config: DictConfig) -> None:
        self.model: Any
        self.model_path: str | None
        self.device: str
        self.config = config
        self.is_loaded = False
        self.model_dtype: np.dtype | torch.dtype

    @abstractmethod
    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        pass

    @abstractmethod
    def load_from_disk(self, model_path: str, device: str = "cpu"):
        """Load model from disk."""

    @abstractmethod
    def to(self, device: str):
        """Move model to device."""

    @abstractmethod
    def eval(self):
        """Set model to evaluation mode."""

    @abstractmethod
    def half(self):
        """Convert model to half precision."""

    @abstractmethod
    def cpu(self):
        """Move model to cpu."""

    @property
    def training(self) -> bool:
        """Return whether model is in training mode."""
        return False

    @property
    def device(self) -> str:
        """Return the device of the model."""
        return self._device

    @device.setter
    def device(self, device: str):
        """Set the device of the model."""
        if device == "cuda" and ":" not in device:
            device = f"{device}:0"

        self._device = device


class TorchscriptEvaluationModel(BaseEvaluationModel):
    """Wrapper for torchscript models."""

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return self.model(*args, **kwargs)

    def load_from_disk(self, model_path: str, device: str = "cpu"):
        """Load model from disk."""
        self.model_path = model_path
        self.device = device

        model = cast(RecursiveScriptModule, torch.jit.load(self.model_path))
        model.eval()
        model.to(self.device)

        parameter_types = {param.dtype for param in model.parameters()}
        if len(parameter_types) == 2:
            # TODO: There could be models with mixed precision?
            raise ValueError(f"Expected only one type of parameters, found {parameter_types}")

        self.model_dtype = list(parameter_types)[0]
        self.model = model
        self.is_loaded = True

    def to(self, device: str):
        """Move model to device."""
        self.model.to(device)
        self.device = device

    def eval(self):
        """Set model to evaluation mode."""
        self.model.eval()

    @property
    def training(self) -> bool:
        """Return whether model is in training mode."""
        return self.model.training

    def half(self):
        """Convert model to half precision."""
        self.model.half()

    def cpu(self):
        """Move model to cpu."""
        self.model.cpu()


class TorchEvaluationModel(TorchscriptEvaluationModel):
    """Wrapper for torch models.

    Args:
        model_architecture: Optional torch model architecture
    """

    def __init__(self, config: DictConfig, model_architecture: nn.Module) -> None:
        super().__init__(config=config)
        self.model = model_architecture
        self.model.eval()
        device = next(self.model.parameters()).device
        self.device = str(device)

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return self.model(*args, **kwargs)

    def load_from_disk(self, model_path: str, device: str = "cpu"):
        """Load model from disk."""
        self.model_path = model_path
        self.device = device
        self.model.load_state_dict(torch.load(self.model_path))
        self.model.eval()
        self.model.to(self.device)

        parameter_types = {param.dtype for param in self.model.parameters()}
        if len(parameter_types) == 2:
            # TODO: There could be models with mixed precision?
            raise ValueError(f"Expected only one type of parameters, found {parameter_types}")

        self.model_dtype = list(parameter_types)[0]
        self.is_loaded = True


onnx_to_torch_dtype_dict = {
    "tensor(bool)": torch.bool,
    "tensor(uint8)": torch.uint8,
    "tensor(int8)": torch.int8,
    "tensor(int16)": torch.int16,
    "tensor(int32)": torch.int32,
    "tensor(int64)": torch.int64,
    "tensor(float16)": torch.float16,
    "tensor(float32)": torch.float32,
    "tensor(float)": torch.float32,
    "tensor(float64)": torch.float64,
    "tensor(complex64)": torch.complex64,
    "tensor(complex128)": torch.complex128,
}


class ONNXEvaluationModel(BaseEvaluationModel):
    """Wrapper for ONNX models. It's designed to provide a similar interface to standard torch models."""

    def __init__(self, config: DictConfig) -> None:
        if not ONNX_AVAILABLE:
            raise ImportError(
                "onnxruntime is not installed. Please install ONNX capabilities for quadra with: poetry install -E onnx"
            )
        super().__init__(config=config)
        self.session_options = self.generate_session_options()

    def generate_session_options(self) -> ort.SessionOptions:
        """Generate session options from the current config."""
        session_options = ort.SessionOptions()

        if hasattr(self.config, "session_options") and self.config.session_options is not None:
            session_options_dict = cast(
                dict[str, Any], OmegaConf.to_container(self.config.session_options, resolve=True)
            )
            for key, value in session_options_dict.items():
                final_value = value
                if isinstance(value, dict) and "_target_" in value:
                    final_value = instantiate(final_value)

                setattr(session_options, key, final_value)

        return session_options

    def __call__(self, *inputs: np.ndarray | torch.Tensor) -> Any:
        """Run inference on the model and return the output as torch tensors."""
        # TODO: Maybe we can support also kwargs
        use_pytorch = False

        onnx_inputs: dict[str, np.ndarray | torch.Tensor] = {}

        for onnx_input, current_input in zip(self.model.get_inputs(), inputs, strict=False):
            if isinstance(current_input, torch.Tensor):
                onnx_inputs[onnx_input.name] = current_input
                use_pytorch = True
            elif isinstance(current_input, np.ndarray):
                onnx_inputs[onnx_input.name] = current_input
            else:
                raise ValueError(f"Invalid input type: {type(inputs)}")

            if use_pytorch and isinstance(current_input, np.ndarray):
                raise ValueError("Cannot mix torch and numpy inputs")

        if use_pytorch:
            onnx_output = self._forward_from_pytorch(cast(dict[str, torch.Tensor], onnx_inputs))
        else:
            onnx_output = self._forward_from_numpy(cast(dict[str, np.ndarray], onnx_inputs))

        onnx_output = [torch.from_numpy(x).to(self.device) if isinstance(x, np.ndarray) else x for x in onnx_output]

        if len(onnx_output) == 1:
            onnx_output = onnx_output[0]

        return onnx_output

    def _forward_from_pytorch(self, input_dict: dict[str, torch.Tensor]):
        """Run inference on the model and return the output as torch tensors."""
        io_binding = self.model.io_binding()
        device_type = self.device.split(":")[0]

        for k, v in input_dict.items():
            if not v.is_contiguous():
                # If not contiguous onnx give wrong results
                v = v.contiguous()  # noqa: PLW2901

            io_binding.bind_input(
                name=k,
                device_type=device_type,
                # Weirdly enough onnx wants 0 for cpu
                device_id=0 if device_type == "cpu" else int(self.device.split(":")[1]),
                element_type=np.float16 if v.dtype == torch.float16 else np.float32,
                shape=tuple(v.shape),
                buffer_ptr=v.data_ptr(),
            )

        for x in self.model.get_outputs():
            # TODO: Is it possible to also bind the output? We require info about output dimensions
            io_binding.bind_output(name=x.name)

        self.model.run_with_iobinding(io_binding)

        output = io_binding.copy_outputs_to_cpu()

        return output

    def _forward_from_numpy(self, input_dict: dict[str, np.ndarray]):
        """Run inference on the model and return the output as numpy array."""
        ort_outputs = [x.name for x in self.model.get_outputs()]

        onnx_output = self.model.run(ort_outputs, input_dict)

        return onnx_output

    def load_from_disk(self, model_path: str, device: str = "cpu"):
        """Load model from disk."""
        self.model_path = model_path
        self.device = device

        ort_providers = self._get_providers(device)
        self.model = ort.InferenceSession(self.model_path, providers=ort_providers, sess_options=self.session_options)
        self.model_dtype = self.cast_onnx_dtype(self.model.get_inputs()[0].type)
        self.is_loaded = True

    def _get_providers(self, device: str) -> list[tuple[str, dict[str, Any]] | str]:
        """Return the providers for the ONNX model based on the device."""
        ort_providers: list[tuple[str, dict[str, Any]] | str]

        if device == "cpu":
            ort_providers = ["CPUExecutionProvider"]
        else:
            ort_providers = [
                (
                    "CUDAExecutionProvider",
                    {
                        "device_id": int(device.split(":")[1]),
                    },
                )
            ]

        return ort_providers

    def to(self, device: str):
        """Move model to device."""
        self.device = device
        ort_providers = self._get_providers(device)
        self.model.set_providers(ort_providers)

    def eval(self):
        """Fake interface to match torch models."""
        return self

    def half(self):
        """Convert model to half precision."""
        raise NotImplementedError("At the moment ONNX models do not support half method.")

    def cpu(self):
        """Move model to cpu."""
        self.to("cpu")

    def cast_onnx_dtype(self, onnx_dtype: str) -> torch.dtype | np.dtype:
        """Cast ONNX dtype to numpy or pytorch dtype."""
        return onnx_to_torch_dtype_dict[onnx_dtype]



from __future__ import annotations

import os

import cv2
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.cm import get_cmap
from matplotlib.colors import Colormap
from matplotlib.lines import Line2D
from matplotlib.pyplot import Figure

from quadra.utils import utils

log = utils.get_logger(__name__)


def plot_patch_reconstruction(
    reconstruction: dict,
    idx_to_class: dict[int, str],
    class_to_idx: dict[str, int],
    ignore_classes: list[int] | None = None,
    is_polygon: bool = True,
) -> Figure:
    """Helper function for plotting the patch reconstruction.

    Args:
        reconstruction: Dict following this structure
            {
                "file_path": str,
                "mask_path": str,
                "prediction": {
                    "label": str,
                    "points": [{"x": int, "y": int}]
                }
            } if is_polygon else
            {
                "file_path": str,
                "mask_path": str,
                "prediction": np.ndarray
            }
        idx_to_class: Dictionary mapping indices to label names
        class_to_idx: Dictionary mapping class names to indices
        ignore_classes: Eventually the classes to not plot
        is_polygon: Boolean indicating if the prediction is a polygon or a mask.

    Returns:
        Matplotlib plot showing predicted patch regions and eventually gt

    """
    cmap_name = "tab10"

    # 10 classes + good
    if len(idx_to_class.values()) > 11:
        cmap_name = "tab20"

    cmap = get_cmap(cmap_name)
    test_img = cv2.imread(reconstruction["image_path"])
    test_img = cv2.cvtColor(test_img, cv2.COLOR_BGR2RGB)
    gt_img = None

    if reconstruction["mask_path"] is not None and os.path.isfile(reconstruction["mask_path"]):
        gt_img = cv2.imread(reconstruction["mask_path"], 0)

    out = np.zeros((test_img.shape[0], test_img.shape[1]), dtype=np.uint8)

    if is_polygon:
        for _, region in enumerate(reconstruction["prediction"]):
            points = [[item["x"], item["y"]] for item in region["points"]]
            c_label = region["label"]

            out = cv2.drawContours(  # type: ignore[call-overload]
                out,
                np.array([points], np.int32),
                -1,
                class_to_idx[c_label],
                thickness=cv2.FILLED,
            )
    else:
        out = reconstruction["prediction"]

    fig = plot_patch_results(
        image=test_img,
        prediction_image=out,
        ground_truth_image=gt_img,
        plot_original=True,
        ignore_classes=ignore_classes,
        save_path=None,
        class_to_idx=class_to_idx,
        cmap=cmap,
    )

    return fig


def show_mask_on_image(image: np.ndarray, mask: np.ndarray):
    """Plot mask on top of the original image."""
    image = image.astype(np.float32) / 255
    mask = mask.astype(np.float32) / 255
    out = mask + image.astype(np.float32)
    out = out / np.max(out)
    return np.uint8(255 * out)


def create_rgb_mask(
    mask: np.ndarray,
    color_map: dict,
    ignore_classes: list[int] | None = None,
    ground_truth_mask: np.ndarray | None = None,
):
    """Convert index mask to RGB mask."""
    output_mask = np.zeros([mask.shape[0], mask.shape[1], 3])
    for c in np.unique(mask):
        if ignore_classes is not None and c in ignore_classes:
            continue

        output_mask[mask == c] = color_map[str(c)]
    if ignore_classes is not None and ground_truth_mask is not None:
        output_mask[np.isin(ground_truth_mask, ignore_classes)] = [0, 0, 0]

    return output_mask


def plot_patch_results(
    image: np.ndarray,
    prediction_image: np.ndarray,
    ground_truth_image: np.ndarray | None,
    class_to_idx: dict[str, int],
    plot_original: bool = True,
    ignore_classes: list[int] | None = None,
    image_height: int = 10,
    save_path: str | None = None,
    cmap: Colormap | None = None,
) -> Figure:
    """Function used to plot the image predicted.

    Args:
        prediction_image: The prediction image
        image: The original image to plot
        ground_truth_image: The ground truth image
        class_to_idx: Dictionary mapping class names to indices
        plot_original: Boolean to plot the original image
        ignore_classes: The classes to ignore, default is 0
        image_height: The height of the output figure
        save_path: The path to save the figure
        cmap: The colormap to use. If None, tab20 is used

    Returns:
        The matplotlib figure
    """
    if ignore_classes is None:
        ignore_classes = [0]

    if cmap is None:
        cmap = get_cmap("tab20")

    image = image[0 : prediction_image.shape[0], 0 : prediction_image.shape[1], :]
    idx_to_class = {v: k for k, v in class_to_idx.items()}

    if ignore_classes is not None:
        class_to_idx = {k: v for k, v in class_to_idx.items() if v not in ignore_classes}

    class_idxs = list(class_to_idx.values())

    cmap = {str(c): tuple(int(i * 255) for i in cmap(c / len(class_idxs))[:-1]) for c in class_idxs}
    output_images = []
    titles = []

    if plot_original:
        output_images.append(image)
        titles.append("Original Image")

    if ground_truth_image is not None:
        ground_truth_image = ground_truth_image[0 : prediction_image.shape[0], 0 : prediction_image.shape[1]]
        ground_truth_mask = create_rgb_mask(ground_truth_image, cmap, ignore_classes=ignore_classes)
        output_images.append(ground_truth_mask)
        titles.append("Ground Truth Mask")

    prediction_mask = create_rgb_mask(
        prediction_image,
        cmap,
        ignore_classes=ignore_classes,
    )

    output_images.append(prediction_mask)
    titles.append("Prediction Mask")
    if ignore_classes is not None and ground_truth_image is not None:
        prediction_mask = create_rgb_mask(
            prediction_image, cmap, ignore_classes=ignore_classes, ground_truth_mask=ground_truth_image
        )

        ignored_classes_str = [idx_to_class[c] for c in ignore_classes]
        prediction_title = f"Prediction Mask \n (Ignoring Ground Truth Class: {ignored_classes_str})"
        output_images.append(prediction_mask)
        titles.append(prediction_title)

    fig, axs = plt.subplots(
        ncols=len(output_images),
        nrows=1,
        figsize=(len(output_images) * image_height, image_height),
        squeeze=False,
        facecolor="white",
    )

    for i, output_image in enumerate(output_images):
        axs[0, i].imshow(show_mask_on_image(image, output_image))
        axs[0, i].set_title(titles[i])
        axs[0, i].axis("off")

    custom_lines = [Line2D([0], [0], color=tuple(i / 255.0 for i in cmap[str(c)]), lw=4) for c in class_idxs]
    custom_labels = list(class_to_idx.keys())
    axs[0, -1].legend(custom_lines, custom_labels, loc="center left", bbox_to_anchor=(1.01, 0.81), borderaxespad=0)
    if save_path is not None:
        plt.savefig(save_path, bbox_inches="tight")
        plt.close()

    return fig



import logging

import numpy as np
import torch
import torch.nn.functional as F
from torch import nn

log = logging.getLogger(__name__)


def dino_distillation_loss(
    student_output: torch.Tensor,
    teacher_output: torch.Tensor,
    center_vector: torch.Tensor,
    teacher_temp: float = 0.04,
    student_temp: float = 0.1,
) -> torch.Tensor:
    """Compute the DINO distillation loss.

    Args:
        student_output: tensor of the student output
        teacher_output: tensor of the teacher output
        center_vector: center vector of distribution
        teacher_temp: temperature teacher
        student_temp: temperature student.

    Returns:
        The computed loss
    """
    student_temp = [s / student_temp for s in student_output]
    teacher_temp = [(t - center_vector) / teacher_temp for t in teacher_output]

    student_sm = [F.log_softmax(s, dim=-1) for s in student_temp]
    teacher_sm = [F.softmax(t, dim=-1).detach() for t in teacher_temp]

    total_loss = torch.tensor(0.0, device=student_output[0].device)
    n_loss_terms = torch.tensor(0.0, device=student_output[0].device)

    for t_ix, t in enumerate(teacher_sm):
        for s_ix, s in enumerate(student_sm):
            if t_ix == s_ix:
                continue

            loss = torch.sum(-t * s, dim=-1)  # (n_samples,)
            total_loss += loss.mean()  # scalar
            n_loss_terms += 1

    total_loss /= n_loss_terms
    return total_loss


class DinoDistillationLoss(nn.Module):
    """Dino distillation loss module.

    Args:
        output_dim: output dim.
        max_epochs: max epochs.
        warmup_teacher_temp: warmup temperature.
        teacher_temp: teacher temperature.
        warmup_teacher_temp_epochs: warmup teacher epocs.
        student_temp: student temperature.
        center_momentum: center momentum.
    """

    def __init__(
        self,
        output_dim: int,
        max_epochs: int,
        warmup_teacher_temp: float = 0.04,
        teacher_temp: float = 0.07,
        warmup_teacher_temp_epochs: int = 30,
        student_temp: float = 0.1,
        center_momentum: float = 0.9,
    ):
        super().__init__()
        self.student_temp = student_temp
        self.center_momentum = center_momentum
        self.center: torch.Tensor
        # we apply a warm up for the teacher temperature because
        # a too high temperature makes the training instable at the beginning

        if warmup_teacher_temp_epochs >= max_epochs:
            raise ValueError(
                f"Number of warmup epochs ({warmup_teacher_temp_epochs}) must be smaller than max_epochs ({max_epochs})"
            )

        if warmup_teacher_temp_epochs < 30:
            log.warning("Warmup teacher epochs is very small (< 30). This may cause instabilities in the training")

        self.teacher_temp_schedule = np.concatenate(
            (
                np.linspace(warmup_teacher_temp, teacher_temp, warmup_teacher_temp_epochs),
                np.ones(max_epochs - warmup_teacher_temp_epochs) * teacher_temp,
            )
        )
        self.register_buffer("center", torch.zeros(1, output_dim))

    def forward(
        self,
        current_epoch: int,
        student_output: torch.Tensor,
        teacher_output: torch.Tensor,
    ) -> torch.Tensor:
        """Runs forward."""
        teacher_temp = self.teacher_temp_schedule[current_epoch]
        loss = dino_distillation_loss(
            student_output,
            teacher_output,
            center_vector=self.center,
            teacher_temp=teacher_temp,
            student_temp=self.student_temp,
        )

        self.update_center(teacher_output)
        return loss

    @torch.no_grad()
    def update_center(self, teacher_output: torch.Tensor) -> None:
        """Update center of the distribution of the teacher
        Args:
            teacher_output: teacher output.

        Returns:
            None
        """
        # TODO: check if this is correct
        # torch.cat expects a list of tensors but teacher_output is a tensor
        batch_center = torch.cat(teacher_output).mean(dim=0, keepdim=True)  # type: ignore[call-overload]
        self.center = self.center * self.center_momentum + batch_center * (1 - self.center_momentum)



"""Anomaly Score Normalization Callback that uses min-max normalization."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

try:
    from typing import Any, TypeAlias
except ImportError:
    from typing import Any

    from typing_extensions import TypeAlias  # noqa


# MyPy wants TypeAlias, but pylint has problems dealing with it
import numpy as np  # pylint: disable=unused-import
import pytorch_lightning as pl
import torch  # pylint: disable=unused-import
from anomalib.models.components import AnomalyModule
from pytorch_lightning import Callback
from pytorch_lightning.utilities.types import STEP_OUTPUT

# https://github.com/python/cpython/issues/90015#issuecomment-1172996118
MapOrValue: TypeAlias = "float | torch.Tensor | np.ndarray"


def normalize_anomaly_score(raw_score: MapOrValue, threshold: float) -> MapOrValue:
    """Normalize anomaly score value or map based on threshold.

    Args:
        raw_score: Raw anomaly score valure or map
        threshold: Threshold for anomaly detection

    Returns:
        Normalized anomaly score value or map clipped between 0 and 1000
    """
    if threshold > 0:
        normalized_score = (raw_score / threshold) * 100.0
    elif threshold == 0:
        # TODO: Is this the best way to handle this case?
        normalized_score = (raw_score + 1) * 100.0
    else:
        normalized_score = 200.0 - ((raw_score / threshold) * 100.0)

    if isinstance(normalized_score, torch.Tensor):
        return torch.clamp(normalized_score, 0.0, 1000.0)

    return np.clip(normalized_score, 0.0, 1000.0)


class ThresholdNormalizationCallback(Callback):
    """Callback that normalizes the image-level and pixel-level anomaly scores dividing by the threshold value.

    Args:
        threshold_type: Threshold used to normalize pixel level anomaly scores, either image or pixel (default)
    """

    def __init__(self, threshold_type: str = "pixel"):
        super().__init__()
        self.threshold_type = threshold_type

    def on_test_start(self, trainer: pl.Trainer, pl_module: AnomalyModule) -> None:
        """Called when the test begins."""
        del trainer  # `trainer` variable is not used.

        for metric in (pl_module.image_metrics, pl_module.pixel_metrics):
            if metric is not None:
                metric.set_threshold(100.0)

    def on_test_batch_end(
        self,
        trainer: pl.Trainer,
        pl_module: AnomalyModule,
        outputs: STEP_OUTPUT | None,
        batch: Any,
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> None:
        """Called when the test batch ends, normalizes the predicted scores and anomaly maps."""
        del trainer, batch, batch_idx, dataloader_idx  # These variables are not used.

        self._normalize_batch(outputs, pl_module)

    def on_predict_batch_end(
        self,
        trainer: pl.Trainer,
        pl_module: AnomalyModule,
        outputs: Any,
        batch: Any,
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> None:
        """Called when the predict batch ends, normalizes the predicted scores and anomaly maps."""
        del trainer, batch, batch_idx, dataloader_idx  # These variables are not used.

        self._normalize_batch(outputs, pl_module)

    def _normalize_batch(self, outputs, pl_module):
        """Normalize a batch of predictions."""
        image_threshold = pl_module.image_threshold.value.cpu()
        pixel_threshold = pl_module.pixel_threshold.value.cpu()
        outputs["pred_scores"] = normalize_anomaly_score(outputs["pred_scores"], image_threshold.item())

        threshold = pixel_threshold if self.threshold_type == "pixel" else image_threshold
        threshold = threshold.item()

        if "anomaly_maps" in outputs:
            outputs["anomaly_maps"] = normalize_anomaly_score(outputs["anomaly_maps"], threshold)

        if "box_scores" in outputs:
            outputs["box_scores"] = [normalize_anomaly_score(scores, threshold) for scores in outputs["box_scores"]]



from __future__ import annotations

from collections.abc import Callable
from enum import Enum

import torch
import torch.nn.functional as F
from torch import nn, optim

from quadra.losses.ssl import hyperspherical as loss
from quadra.modules.base import BaseLightningModule


class AlignLoss(Enum):
    """Align loss enum."""

    L2 = 1
    COSINE = 2


class TLHyperspherical(BaseLightningModule):
    """Hyperspherical model: maps features extracted from a pretrained backbone into
    an hypersphere.

    Args:
        model: Feature extractor as pytorch `torch.nn.Module`
        optimizer: optimizer of the training.
            If None a default Adam is used.
        lr_scheduler: lr scheduler.
            If None a default ReduceLROnPlateau is used.
        align_weight: Weight for the align loss component for the
            hyperspherical loss.
            Defaults to 1.
        unifo_weight: Weight for the uniform loss component for the
            hyperspherical loss.
            Defaults to 1.
        classifier_weight: Weight for the classifier loss component for the
            hyperspherical loss.
            Defaults to 1.
        align_loss_type: Which type of align loss to use.
            Defaults to AlignLoss.L2.
        classifier_loss: Whether to compute a classifier loss to 'enhance'
            the hyperpsherical loss with the classification loss.
            It True, model.classifier must be defined
            Defaults to False.
        num_classes: Number of classes for a classification problem.
            Defaults to None.
    """

    def __init__(
        self,
        model: nn.Module,
        optimizer: optim.Optimizer | None = None,
        lr_scheduler: object | None = None,
        align_weight: float = 1,
        unifo_weight: float = 1,
        classifier_weight: float = 1,
        align_loss_type: AlignLoss = AlignLoss.L2,
        classifier_loss: bool = False,
        num_classes: int | None = None,
    ):
        super().__init__(model, optimizer, lr_scheduler)
        self.align_loss_fun: (
            Callable[[torch.Tensor, torch.Tensor, int], torch.Tensor]
            | Callable[[torch.Tensor, torch.Tensor], torch.Tensor]
        )
        self.align_weight = align_weight
        self.unifo_weight = unifo_weight
        self.classifier_weight = classifier_weight
        self.align_loss_type = align_loss_type
        if align_loss_type == AlignLoss.L2:
            self.align_loss_fun = loss.align_loss
        elif align_loss_type == AlignLoss.COSINE:
            self.align_loss_fun = loss.cosine_align_loss
        else:
            raise ValueError("The align loss must be one of 'AlignLoss.L2' (L2 distance) or AlignLoss.COSINE")

        if classifier_loss and model.classifier is None:
            raise AssertionError("Classifier is not defined")

        self.classifier_loss = classifier_loss
        self.num_classes = num_classes

    def forward(self, x):
        return self.model(x)

    def training_step(self, batch, batch_idx):
        # pylint: disable=unused-argument
        im_x, im_y, target = batch
        emb_x, emb_y = self(torch.cat([im_x, im_y])).chunk(2)

        align_loss = 0.0
        if self.align_weight > 0:
            align_loss = self.align_loss_fun(emb_x, emb_y)

        unifo_loss = 0.0
        if self.unifo_weight > 0:
            unifo_loss = (loss.uniform_loss(emb_x) + loss.uniform_loss(emb_y)) / 2

        classifier_loss = 0.0
        if self.classifier_loss:
            pred = self.model.classifier(emb_x)
            classifier_loss = F.cross_entropy(pred, target)

        total_loss = (
            self.align_weight * align_loss + self.unifo_weight * unifo_loss + self.classifier_weight * classifier_loss
        )

        self.log(
            "t_loss",
            total_loss,
            on_epoch=True,
            logger=True,
            prog_bar=False,
        )
        self.log(
            "t_align",
            align_loss,
            on_epoch=True,
            on_step=False,
            logger=True,
            prog_bar=False,
        )
        self.log(
            "t_classifier",
            classifier_loss,
            on_epoch=True,
            on_step=False,
            logger=True,
            prog_bar=True,
        )
        self.log(
            "t_unif",
            unifo_loss,
            on_epoch=True,
            on_step=False,
            logger=True,
            prog_bar=False,
        )
        return {"loss": total_loss}

    def train_epoch_end(self, outputs):
        avg_loss = torch.stack([x["loss"] for x in outputs]).mean()

        return {"loss": avg_loss}

    def validation_step(self, batch, batch_idx):
        # pylint: disable=unused-argument
        im_x, im_y, target = batch
        emb_x, emb_y = self(torch.cat([im_x, im_y])).chunk(2)

        align_loss = 0.0
        if self.align_weight > 0:
            align_loss = self.align_loss_fun(emb_x, emb_y)

        unifo_loss = 0.0
        if self.unifo_weight > 0:
            unifo_loss = (loss.uniform_loss(emb_x) + loss.uniform_loss(emb_y)) / 2

        classifier_loss = 0.0
        if self.classifier_loss:
            pred = self.model.classifier(emb_x)
            classifier_loss = F.cross_entropy(pred, target)

        total_loss = (
            self.align_weight * align_loss + self.unifo_weight * unifo_loss + self.classifier_weight * classifier_loss
        )

        self.log(
            "val_loss",
            total_loss,
            on_epoch=True,
            on_step=False,
            logger=True,
            prog_bar=False,
        )
        self.log(
            "v_classifier",
            classifier_loss,
            on_epoch=True,
            on_step=False,
            logger=True,
            prog_bar=True,
        )
        self.log(
            "v_align",
            align_loss,
            on_epoch=True,
            on_step=False,
            logger=True,
            prog_bar=False,
        )
        self.log(
            "v_unif",
            unifo_loss,
            on_epoch=True,
            on_step=False,
            logger=True,
            prog_bar=False,
        )
        return {"val_loss": total_loss}

    def on_validation_epoch_end(self, outputs):
        avg_loss = torch.stack([x["val_loss"] for x in outputs]).mean()

        return {"val_loss": avg_loss}



from __future__ import annotations

import difflib
import importlib
import inspect
from collections.abc import Iterable
from typing import Any

from omegaconf import DictConfig, ListConfig, OmegaConf

from quadra.utils.utils import get_logger

OMEGACONF_FIELDS: tuple[str, ...] = ("_target_", "_convert_", "_recursive_", "_args_")
EXCLUDE_KEYS: tuple[str, ...] = ("hydra",)

logger = get_logger(__name__)


def get_callable_arguments(full_module_path: str) -> tuple[list[str], bool]:
    """Gets all arguments from module path.

    Args:
        full_module_path:  Full module path to the target class or function.

    Raises:
        ValueError: If the target is not a class or a function.

    Returns:
        All arguments from the target class or function.
    """
    module_path, callable_name = full_module_path.rsplit(".", 1)
    module = importlib.import_module(module_path)
    callable_ = getattr(module, callable_name)
    # check if it is a class
    accepts_kwargs = False
    if inspect.isclass(callable_):
        arg_names = []
        for cls in callable_.__mro__:
            if cls is object:
                break
            # We don' access the instance but mypy complains
            init_argspec = inspect.getfullargspec(cls.__init__)  # type: ignore
            cls_arg_names = init_argspec.args[1:]
            cls_kwonlyargs = init_argspec.kwonlyargs
            arg_names.extend(cls_arg_names)
            arg_names.extend(cls_kwonlyargs)
            # if the target class or function accepts kwargs, we cannot check arguments
            accepts_kwargs = init_argspec.varkw is not None or accepts_kwargs
        arg_names = list(set(arg_names))
    elif inspect.isfunction(callable_):
        init_argspec = inspect.getfullargspec(callable_)
        arg_names = []
        arg_names.extend(init_argspec.args)
        arg_names.extend(init_argspec.kwonlyargs)
        accepts_kwargs = init_argspec.varkw is not None or accepts_kwargs
    else:
        raise ValueError("The target must be a class or a function.")

    return arg_names, accepts_kwargs


def check_all_arguments(callable_variable: str, configuration_arguments: list[str], argument_names: list[str]) -> None:
    """Checks if all arguments passed from configuration are valid for the target class or function.

    Args:
        callable_variable : Full module path to the target class or function.
        configuration_arguments : All arguments passed from configuration.
        argument_names: All arguments from the target class or function.

    Raises:
        ValueError: If the argument is not valid for the target class or function.
    """
    for argument in configuration_arguments:
        if argument not in argument_names:
            error_string = f"`{argument}` is not a valid argument passed from configuration to `{callable_variable}`."
            closest_match = difflib.get_close_matches(argument, argument_names, n=1, cutoff=0.5)
            if len(closest_match) > 0:
                error_string += f" Did you mean `{closest_match[0]}`?"
            raise ValueError(error_string)


def validate_config(_cfg: DictConfig | ListConfig, package_name: str = "quadra") -> None:
    """Recursively traverse OmegaConf object and check if arguments are valid for the target class or function.
    If not, raise a ValueError with a suggestion for the closest match of the argument name.

    Args:
        _cfg: OmegaConf object
        package_name: package name to check for instantiation.
    """
    # The below lines of code for looping over a DictConfig/ListConfig are
    # borrowed from OmegaConf PR #719.
    itr: Iterable[Any]
    if isinstance(_cfg, ListConfig):
        itr = range(len(_cfg))
    else:
        itr = _cfg
    for key in itr:
        if OmegaConf.is_missing(_cfg, key):
            continue
        if isinstance(key, str) and any(x in key for x in EXCLUDE_KEYS):
            continue
        if OmegaConf.is_config(_cfg[key]):
            validate_config(_cfg[key])
        elif isinstance(_cfg[key], str):
            if key == "_target_":
                callable_variable = str(_cfg[key])
                if callable_variable.startswith(package_name):
                    configuration_arguments = [str(x) for x in _cfg if x not in OMEGACONF_FIELDS]
                    argument_names, accepts_kwargs = get_callable_arguments(callable_variable)
                    if not accepts_kwargs:
                        check_all_arguments(callable_variable, configuration_arguments, argument_names)
        else:
            logger.debug("Skipping %s from config. It is not supported.", key)



import torch


def vicreg_loss(
    z1: torch.Tensor,
    z2: torch.Tensor,
    lambd: float,
    mu: float,
    nu: float = 1,
    gamma: float = 1,
) -> torch.Tensor:
    """VICReg loss described in https://arxiv.org/abs/2105.04906.

    Args:
        z1: First `augmented` normalized features (i.e. f(T(x))). The normalization can be obtained with
            z1_norm = (z1 - z1.mean(0)) / z1.std(0)
        z2: Second `augmented` normalized features (i.e. f(T(x))). The normalization can be obtained with
            z2_norm = (z2 - z2.mean(0)) / z2.std(0)
        lambd: lambda multiplier for redundancy term.
        mu: mu multiplier for similarity term.
        nu: nu multiplier for variance term. Default: 1
        gamma: gamma multiplier for covariance term. Default: 1

    Returns:
        VICReg loss
    """
    # Variance loss
    std_z1 = torch.sqrt(z1.var(dim=0) + 0.0001)
    std_z2 = torch.sqrt(z2.var(dim=0) + 0.0001)
    v_z1 = torch.nn.functional.relu(gamma - std_z1).mean()
    v_z2 = torch.nn.functional.relu(gamma - std_z2).mean()
    var_loss = v_z1 + v_z2

    # Similarity loss
    sim_loss = torch.nn.functional.mse_loss(z1, z2)

    # Covariance loss
    n = z1.size(0)
    d = z1.size(1)
    z1 = z1 - z1.mean(dim=0)
    z2 = z2 - z2.mean(dim=0)
    cov_z1 = (z1.T @ z1) / (n - 1)
    cov_z2 = (z2.T @ z2) / (n - 1)
    off_diagonal_cov_z1 = cov_z1.flatten()[:-1].view(d - 1, d + 1)[:, 1:].flatten()
    off_diagonal_cov_z2 = cov_z2.flatten()[:-1].view(d - 1, d + 1)[:, 1:].flatten()
    cov_loss = off_diagonal_cov_z1.pow_(2).sum() / d + off_diagonal_cov_z2.pow_(2).sum() / d

    return lambd * sim_loss + mu * var_loss + nu * cov_loss


class VICRegLoss(torch.nn.Module):
    """VIC regression loss module.

    Args:
        lambd: lambda multiplier for redundancy term.
        mu: mu multiplier for similarity term.
        nu: nu multiplier for variance term. Default: 1.
        gamma: gamma multiplier for covariance term. Default: 1.
    """

    def __init__(
        self,
        lambd: float,
        mu: float,
        nu: float = 1,
        gamma: float = 1,
    ):
        super().__init__()
        self.lambd = lambd
        self.mu = mu
        self.nu = nu
        self.gamma = gamma

    def forward(self, z1: torch.Tensor, z2: torch.Tensor) -> torch.Tensor:
        """Computes VICReg loss."""
        return vicreg_loss(z1, z2, self.lambd, self.mu, self.nu, self.gamma)



"""Common utility functions.
Some of them are mostly based on https://github.com/ashleve/lightning-hydra-template.
"""

from __future__ import annotations

import glob
import json
import logging
import os
import shutil
import subprocess
import sys
import warnings
from collections.abc import Iterable, Iterator, Sequence
from tempfile import TemporaryDirectory
from typing import Any, cast

import cv2
import dotenv
import mlflow
import numpy as np
import pytorch_lightning as pl
import rich.syntax
import rich.tree
import torch
from hydra.core.hydra_config import HydraConfig
from hydra.utils import get_original_cwd
from lightning_fabric.utilities.device_parser import _parse_gpu_ids
from omegaconf import DictConfig, OmegaConf
from pytorch_lightning.loggers import TensorBoardLogger
from pytorch_lightning.utilities import rank_zero_only

import quadra
import quadra.utils.export as quadra_export
from quadra.callbacks.mlflow import get_mlflow_logger
from quadra.utils.mlflow import infer_signature_model

try:
    import onnx  # noqa

    ONNX_AVAILABLE = True
except ImportError:
    ONNX_AVAILABLE = False


IMAGE_EXTENSIONS: list[str] = [".png", ".jpg", ".jpeg", ".bmp", ".tiff", ".tif", ".pbm", ".pgm", ".ppm", ".pxm", ".pnm"]


def get_logger(name=__name__) -> logging.Logger:
    """Initializes multi-GPU-friendly python logger."""
    logger = logging.getLogger(name)

    # this ensures all logging levels get marked with the rank zero decorator
    # otherwise logs would get multiplied for each GPU process in multi-GPU setup
    for level in ("debug", "info", "warning", "error", "exception", "fatal", "critical"):
        setattr(logger, level, rank_zero_only(getattr(logger, level)))

    return logger


def extras(config: DictConfig) -> None:
    """A couple of optional utilities, controlled by main config file:
    - disabling warnings
    - forcing debug friendly configuration
    - verifying experiment name is set when running in experiment mode.
    Modifies DictConfig in place.

    Args:
        config: Configuration composed by Hydra.
    """
    logging.basicConfig()
    logging.getLogger().setLevel(config.core.log_level.upper())

    log = get_logger(__name__)
    config.core.command += " ".join(sys.argv)
    config.core.experiment_path = os.getcwd()

    # disable python warnings if <config.ignore_warnings=True>
    if config.get("ignore_warnings"):
        log.info("Disabling python warnings! <config.ignore_warnings=True>")
        warnings.filterwarnings("ignore")

    # force debugger friendly configuration if <config.trainer.fast_dev_run=True>
    # debuggers don't like GPUs and multiprocessing
    if config.get("trainer") and config.trainer.get("fast_dev_run"):
        log.info("Forcing debugger friendly configuration! <config.trainer.fast_dev_run=True>")
        if config.trainer.get("gpus"):
            config.trainer.devices = 1
            config.trainer.accelerator = "cpu"
            config.trainer.gpus = None
        if config.datamodule.get("pin_memory"):
            config.datamodule.pin_memory = False
        if config.datamodule.get("num_workers"):
            config.datamodule.num_workers = 0


@rank_zero_only
def print_config(
    config: DictConfig,
    fields: Sequence[str] = (
        "task",
        "trainer",
        "model",
        "datamodule",
        "callbacks",
        "logger",
        "core",
        "backbone",
        "transforms",
        "optimizer",
        "scheduler",
    ),
    resolve: bool = True,
) -> None:
    """Prints content of DictConfig using Rich library and its tree structure.

    Args:
        config: Configuration composed by Hydra.
        fields: Determines which main fields from config will
            be printed and in what order.
        resolve: Whether to resolve reference fields of DictConfig.
    """
    style = "dim"
    tree = rich.tree.Tree("CONFIG", style=style, guide_style=style)

    for field in fields:
        branch = tree.add(field, style=style, guide_style=style)

        config_section = config.get(field)
        branch_content = str(config_section)
        if isinstance(config_section, DictConfig):
            branch_content = OmegaConf.to_yaml(config_section, resolve=resolve)

        branch.add(rich.syntax.Syntax(branch_content, "yaml"))

    rich.print(tree)

    with open("config_tree.txt", "w") as fp:
        rich.print(tree, file=fp)


@rank_zero_only
def log_hyperparameters(
    config: DictConfig,
    model: pl.LightningModule,
    trainer: pl.Trainer,
) -> None:
    """This method controls which parameters from Hydra config are saved by Lightning loggers.

    Additionaly saves:
        - number of trainable model parameters
    """
    log = get_logger(__name__)

    if not HydraConfig.initialized() or trainer.logger is None:
        return

    log.info("Logging hyperparameters!")
    hydra_cfg = HydraConfig.get()
    hydra_choices = OmegaConf.to_container(hydra_cfg.runtime.choices)
    if isinstance(hydra_choices, dict):
        # For multirun override the choices that are not automatically updated
        for item in hydra_cfg.overrides.task:
            if "." in item:
                continue

            override, value = item.split("=")
            hydra_choices[override] = value

        hparams = {}
        hydra_choices_final = {}
        for k, v in hydra_choices.items():
            if isinstance(k, str):
                k_replaced = k.replace("@", "_at_")
                hydra_choices_final[k_replaced] = v
                if v is not None and isinstance(v, str) and "@" in v:
                    hydra_choices_final[k_replaced] = v.replace("@", "_at_")

        hparams.update(hydra_choices_final)
    else:
        logging.warning("Hydra choices is not a dictionary, skip adding them to the logger")
    # save number of model parameters
    hparams["model/params_total"] = sum(p.numel() for p in model.parameters())
    hparams["model/params_trainable"] = sum(p.numel() for p in model.parameters() if p.requires_grad)
    hparams["model/params_not_trainable"] = sum(p.numel() for p in model.parameters() if not p.requires_grad)
    hparams["experiment_path"] = config.core.experiment_path
    hparams["command"] = config.core.command
    hparams["library/version"] = str(quadra.__version__)

    with open(os.devnull, "w") as fnull:
        if subprocess.call(["git", "-C", get_original_cwd(), "status"], stderr=subprocess.STDOUT, stdout=fnull) == 0:
            try:
                hparams["git/commit"] = subprocess.check_output(["git", "rev-parse", "HEAD"]).decode("ascii").strip()
                hparams["git/branch"] = (
                    subprocess.check_output(["git", "rev-parse", "--abbrev-ref", "HEAD"]).decode("ascii").strip()
                )
                hparams["git/remote"] = (
                    subprocess.check_output(["git", "remote", "get-url", "origin"]).decode("ascii").strip()
                )
            except subprocess.CalledProcessError:
                log.warning(
                    "Could not get git commit, branch or remote information, the repository might not have any commits "
                    " yet or it might have been initialized wrongly."
                )
        else:
            log.warning("Could not find git repository, skipping git commit and branch info")

    # send hparams to all loggers
    trainer.logger.log_hyperparams(hparams)


def upload_file_tensorboard(file_path: str, tensorboard_logger: TensorBoardLogger) -> None:
    """Upload a file to tensorboard handling different extensions.

    Args:
        file_path: Path to the file to upload.
        tensorboard_logger: Tensorboard logger instance.
    """
    tag = os.path.basename(file_path)
    ext = os.path.splitext(file_path)[1].lower()

    if ext == ".json":
        with open(file_path) as f:
            json_content = json.load(f)

            json_content = f"```json\n{json.dumps(json_content, indent=4)}\n```"
            tensorboard_logger.experiment.add_text(tag=tag, text_string=json_content, global_step=0)
    elif ext in [".yaml", ".yml"]:
        with open(file_path) as f:
            yaml_content = f.read()
            yaml_content = f"```yaml\n{yaml_content}\n```"
            tensorboard_logger.experiment.add_text(tag=tag, text_string=yaml_content, global_step=0)
    else:
        with open(file_path, encoding="utf-8") as f:
            tensorboard_logger.experiment.add_text(tag=tag, text_string=f.read().replace("\n", "  \n"), global_step=0)

    tensorboard_logger.experiment.flush()


def finish(
    config: DictConfig,
    module: pl.LightningModule,
    datamodule: pl.LightningDataModule,
    trainer: pl.Trainer,
    callbacks: list[pl.Callback],
    logger: list[pl.loggers.Logger],
    export_folder: str,
) -> None:
    """Upload config files to MLFlow server.

    Args:
        config: Configuration composed by Hydra.
        module: LightningModule.
        datamodule: LightningDataModule.
        trainer: LightningTrainer.
        callbacks: List of LightningCallbacks.
        logger: List of LightningLoggers.
        export_folder: Folder where the deployment models are exported.
    """
    # pylint: disable=unused-argument
    if len(logger) > 0 and config.core.get("upload_artifacts"):
        mlflow_logger = get_mlflow_logger(trainer=trainer)
        tensorboard_logger = get_tensorboard_logger(trainer=trainer)
        file_names = ["config.yaml", "config_resolved.yaml", "config_tree.txt", "data/dataset.csv"]
        if "16" in str(trainer.precision):
            index = _parse_gpu_ids(config.trainer.devices, include_cuda=True)[0]
            device = "cuda:" + str(index)
            half_precision = True
        else:
            device = "cpu"
            half_precision = False

        if mlflow_logger is not None:
            config_paths = []

            for f in file_names:
                if os.path.isfile(os.path.join(os.getcwd(), f)):
                    config_paths.append(os.path.join(os.getcwd(), f))

            for path in config_paths:
                mlflow_logger.experiment.log_artifact(
                    run_id=mlflow_logger.run_id, local_path=path, artifact_path="metadata"
                )

            deployed_models = glob.glob(os.path.join(export_folder, "*"))
            model_json: dict[str, Any] | None = None

            if os.path.exists(os.path.join(export_folder, "model.json")):
                with open(os.path.join(export_folder, "model.json")) as json_file:
                    model_json = json.load(json_file)

            if model_json is not None:
                input_size = model_json["input_size"]
                # Not a huge fan of this check
                if not isinstance(input_size[0], list):
                    # Input size is not a list of lists
                    input_size = [input_size]
                inputs = cast(
                    list[Any],
                    quadra_export.generate_torch_inputs(input_size, device=device, half_precision=half_precision),
                )
                types_to_upload = config.core.get("upload_models")
                mlflow_zip_models = config.core.get("mlflow_zip_models", False)
                model_uploaded = False
                with mlflow.start_run(run_id=mlflow_logger.run_id) as _:
                    for model_path in deployed_models:
                        model_type = model_type_from_path(model_path)
                        model_name = os.path.basename(model_path)

                        if model_type is None:
                            logging.warning("%s model type not supported", model_path)
                            continue
                        if model_type is not None and model_type in types_to_upload:
                            if model_type == "pytorch" and not mlflow_zip_models:
                                logging.warning("Pytorch format still not supported for mlflow upload")
                                continue

                            if mlflow_zip_models:
                                with TemporaryDirectory() as temp_dir:
                                    if model_type == "pytorch" and os.path.isfile(
                                        os.path.join(export_folder, "model_config.yaml")
                                    ):
                                        shutil.copy(model_path, temp_dir)
                                        shutil.copy(os.path.join(export_folder, "model_config.yaml"), temp_dir)
                                        shutil.make_archive("assets", "zip", root_dir=temp_dir)
                                    else:
                                        shutil.make_archive(
                                            "assets",
                                            "zip",
                                            root_dir=os.path.dirname(model_path),
                                            base_dir=model_name,
                                        )
                                    shutil.move("assets.zip", temp_dir)
                                    mlflow.pyfunc.log_model(
                                        artifact_path=model_path,
                                        loader_module="not.used",
                                        data_path=os.path.join(temp_dir, "assets.zip"),
                                        pip_requirements=[""],
                                    )
                                    model_uploaded = True
                            else:
                                model = quadra_export.import_deployment_model(
                                    model_path,
                                    device=device,
                                    inference_config=config.inference,
                                )

                                if model_type in ["torchscript", "pytorch"]:
                                    signature = infer_signature_model(model.model, inputs)
                                    mlflow.pytorch.log_model(
                                        model.model,
                                        artifact_path=model_path,
                                        signature=signature,
                                    )
                                    model_uploaded = True

                                elif model_type in ["onnx", "simplified_onnx"] and ONNX_AVAILABLE:
                                    if model.model_path is None:
                                        logging.warning(
                                            "Cannot log onnx model on mlflow, \
                                            BaseEvaluationModel 'model_path' attribute is None"
                                        )
                                    else:
                                        signature = infer_signature_model(model, inputs)
                                        model_proto = onnx.load(model.model_path)
                                        mlflow.onnx.log_model(
                                            model_proto,
                                            artifact_path=model_path,
                                            signature=signature,
                                        )
                                        model_uploaded = True

                    if model_uploaded:
                        mlflow.log_artifact(os.path.join(export_folder, "model.json"), export_folder)

        if tensorboard_logger is not None:
            config_paths = []
            for f in file_names:
                if os.path.isfile(os.path.join(os.getcwd(), f)):
                    config_paths.append(os.path.join(os.getcwd(), f))

            for path in config_paths:
                upload_file_tensorboard(file_path=path, tensorboard_logger=tensorboard_logger)

            tensorboard_logger.experiment.flush()


def load_envs(env_file: str | None = None) -> None:
    """Load all the environment variables defined in the `env_file`.
    This is equivalent to `. env_file` in bash.

    It is possible to define all the system specific variables in the `env_file`.

    Args:
        env_file: the file that defines the environment variables to use. If None
                     it searches for a `.env` file in the project.
    """
    dotenv.load_dotenv(dotenv_path=env_file, override=True)


def model_type_from_path(model_path: str) -> str | None:
    """Determine the type of the machine learning model based on its file extension.

    Parameters:
    - model_path (str): The file path of the machine learning model.

    Returns:
    - str: The type of the model, which can be one of the following:
      - "torchscript" if the model has a '.pt' extension (TorchScript).
      - "pytorch" if the model has a '.pth' extension (PyTorch).
      - "simplified_onnx" if the model file ends with 'simplified.onnx' (Simplified ONNX).
      - "onnx" if the model has a '.onnx' extension (ONNX).
      - "json" if the model has a '.json' extension (JSON).
      - None if model extension is not supported.

    Example:
    ```python
    model_path = "path/to/your/model.onnx"
    model_type = model_type_from_path(model_path)
    print(f"The model type is: {model_type}")
    ```
    """
    if model_path.endswith(".pt"):
        return "torchscript"
    if model_path.endswith(".pth"):
        return "pytorch"
    if model_path.endswith("simplified.onnx"):
        return "simplified_onnx"
    if model_path.endswith(".onnx"):
        return "onnx"
    if model_path.endswith(".json"):
        return "json"
    return None


def setup_opencv() -> None:
    """Setup OpenCV to use only one thread and not use OpenCL."""
    cv2.setNumThreads(1)
    cv2.ocl.setUseOpenCL(False)


def get_device(cuda: bool = True) -> str:
    """Returns the device to use for training.

    Args:
        cuda: whether to use cuda or not

    Returns:
        The device to use
    """
    if torch.cuda.is_available() and cuda:
        return "cuda:0"

    return "cpu"


def nested_set(dic: dict, keys: list[str], value: str) -> None:
    """Assign the value of a dictionary using nested keys."""
    for key in keys[:-1]:
        dic = dic.setdefault(key, {})

    dic[keys[-1]] = value


def flatten_list(input_list: Iterable[Any]) -> Iterator[Any]:
    """Return an iterator over the flattened list.

    Args:
        input_list: the list to be flattened

    Yields:
        The iterator over the flattend list
    """
    for v in input_list:
        if isinstance(v, Iterable) and not isinstance(v, str | bytes):
            yield from flatten_list(v)
        else:
            yield v


class HydraEncoder(json.JSONEncoder):
    """Custom JSON encoder to handle OmegaConf objects."""

    def default(self, o):
        """Convert OmegaConf objects to base python objects."""
        if o is not None and OmegaConf.is_config(o):
            return OmegaConf.to_container(o)
        return json.JSONEncoder.default(self, o)


class NumpyEncoder(json.JSONEncoder):
    """Custom JSON encoder to handle numpy objects."""

    def default(self, o):
        """Custom JSON encoder to handle numpy objects."""
        if o is not None:
            if isinstance(o, np.ndarray):
                if o.size == 1:
                    return o.item()
                return o.tolist()
            if isinstance(o, np.number):
                return o.item()
        return json.JSONEncoder.default(self, o)


class AllGatherSyncFunction(torch.autograd.Function):
    """Function to gather gradients from multiple GPUs."""

    @staticmethod
    def forward(ctx, tensor):
        ctx.batch_size = tensor.shape[0]

        gathered_tensor = [torch.zeros_like(tensor) for _ in range(torch.distributed.get_world_size())]

        torch.distributed.all_gather(gathered_tensor, tensor)
        gathered_tensor = torch.cat(gathered_tensor, 0)

        return gathered_tensor

    @staticmethod
    def backward(ctx, grad_output):
        grad_input = grad_output.clone()
        torch.distributed.all_reduce(grad_input, op=torch.distributed.ReduceOp.SUM, async_op=False)

        idx_from = torch.distributed.get_rank() * ctx.batch_size
        idx_to = (torch.distributed.get_rank() + 1) * ctx.batch_size
        return grad_input[idx_from:idx_to]


@torch.no_grad()
def concat_all_gather(tensor):
    """Performs all_gather operation on the provided tensors.
    *** Warning ***: torch.distributed.all_gather has no gradient.
    """
    tensors_gather = [torch.ones_like(tensor) for _ in range(torch.distributed.get_world_size())]
    torch.distributed.all_gather(tensors_gather, tensor, async_op=False)

    output = torch.cat(tensors_gather, dim=0)
    return output


def get_tensorboard_logger(trainer: pl.Trainer) -> TensorBoardLogger | None:
    """Safely get tensorboard logger from Lightning Trainer loggers.

    Args:
        trainer: Pytorch Lightning Trainer.

    Returns:
        An mlflow logger if available, else None.
    """
    if isinstance(trainer.logger, TensorBoardLogger):
        return trainer.logger

    if isinstance(trainer.logger, list):
        for logger in trainer.logger:
            if isinstance(logger, TensorBoardLogger):
                return logger

    return None



from __future__ import annotations

from collections.abc import Callable
from typing import Any

import torch
from torch.nn import Parameter


class SAM(torch.optim.Optimizer):
    """PyTorch implementation of Sharpness-Aware-Minization paper: https://arxiv.org/abs/2010.01412
    and https://arxiv.org/abs/2102.11600.
    Taken from: https://github.com/davda54/sam.

    Args:
        params: model parameters.
        base_optimizer: optimizer to use.
        rho: Postive float value used to scale the gradients.
        adaptive: Boolean flag indicating whether to use adaptive step update.
        **kwargs: Additional parameters for the base optimizer.
    """

    def __init__(
        self,
        params: list[Parameter],
        base_optimizer: torch.optim.Optimizer,
        rho: float = 0.05,
        adaptive: bool = True,
        **kwargs: Any,
    ):
        assert rho >= 0.0, f"Invalid rho, should be non-negative: {rho}"

        defaults = {"rho": rho, "adaptive": adaptive, **kwargs}
        super().__init__(params, defaults)

        if callable(base_optimizer):
            self.base_optimizer = base_optimizer(self.param_groups, **kwargs)
        else:
            self.base_optimizer = base_optimizer
        self.rho = rho
        self.adaptive = adaptive
        self.param_groups = self.base_optimizer.param_groups

    @torch.no_grad()
    def first_step(self, zero_grad: bool = False) -> None:
        """First step for SAM optimizer.

        Args:
            zero_grad: Boolean flag indicating whether to zero the gradients.

        Returns:
            None
        """
        grad_norm = self._grad_norm()
        for group in self.param_groups:
            scale = self.rho / (grad_norm + 1e-12)

            for p in group["params"]:
                if p.grad is None:
                    continue
                e_w = (torch.pow(p, 2) if self.adaptive else 1.0) * p.grad * scale.to(p)
                p.add_(e_w)  # climb to the local maximum "w + e(w)"
                self.state[p]["e_w"] = e_w

        if zero_grad:
            self.zero_grad()

    @torch.no_grad()
    def second_step(self, zero_grad: bool = False) -> None:
        """Second step for SAM optimizer.

        Args:
            zero_grad: Boolean flag indicating whether to zero the gradients.

        Returns:
            None

        """
        for group in self.param_groups:
            for p in group["params"]:
                if p.grad is None:
                    continue
                p.sub_(self.state[p]["e_w"])  # get back to "w" from "w + e(w)"

        self.base_optimizer.step()  # do the actual "sharpness-aware" update

        if zero_grad:
            self.zero_grad()

    @torch.no_grad()
    def step(self, closure: Callable | None = None) -> None:  # type: ignore[override]
        """Step for SAM optimizer.

        Args:
            closure: The Optional closure for enable grad.

        Returns:
            None

        """
        if closure is not None:
            closure = torch.enable_grad()(closure)

        self.first_step(zero_grad=True)
        if closure is not None:
            closure()
        self.second_step(zero_grad=False)

    def _grad_norm(self) -> torch.Tensor:
        """Put everything on the same device, in case of model parallelism
        Returns:
            Grad norm.
        """
        # put everything on the same device, in case of model parallelism
        shared_device = self.param_groups[0]["params"][0].device
        norm = torch.norm(
            torch.stack(
                [
                    ((torch.abs(p) if self.adaptive else 1.0) * p.grad).norm(p=2).to(shared_device)
                    for group in self.param_groups
                    for p in group["params"]
                    if p.grad is not None
                ]
            ),
            p=2,
        )
        return norm



from __future__ import annotations

import multiprocessing as mp
import multiprocessing.pool as mpp
import os
import pickle as pkl
import typing
from collections.abc import Callable, Iterable, Sequence
from functools import wraps
from typing import Any, Literal, cast

import albumentations
import numpy as np
import pandas as pd
import torch
import xxhash
from pytorch_lightning import LightningDataModule
from tqdm import tqdm

from quadra.utils import utils

log = utils.get_logger(__name__)
TrainDataset = torch.utils.data.Dataset | Sequence[torch.utils.data.Dataset]
ValDataset = torch.utils.data.Dataset | Sequence[torch.utils.data.Dataset]
TestDataset = torch.utils.data.Dataset


def load_data_from_disk_dec(func):
    """Load data from disk if it exists."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        """Wrapper function to load data from disk if it exists."""
        self = cast(BaseDataModule, args[0])
        self.restore_checkpoint()
        return func(*args, **kwargs)

    return wrapper


class DecorateParentMethod(type):
    """Metaclass to decorate methods of subclasses."""

    def __new__(cls, name, bases, dct):
        """Create new  decorator for parent class methods."""
        method_decorator_mapper = {
            "setup": load_data_from_disk_dec,
        }
        for method_name, decorator in method_decorator_mapper.items():
            if method_name in dct:
                dct[method_name] = decorator(dct[method_name])

        return super().__new__(cls, name, bases, dct)


def compute_file_content_hash(path: str, hash_size: Literal[32, 64, 128] = 64) -> str:
    """Get hash of a file based on its content.

    Args:
        path: Path to the file.
        hash_size: Size of the hash. Must be one of [32, 64, 128].

    Returns:
        The hash of the file.
    """
    with open(path, "rb") as f:
        data = f.read()

        if hash_size == 32:
            file_hash = xxhash.xxh32(data, seed=42).hexdigest()
        elif hash_size == 64:
            file_hash = xxhash.xxh64(data, seed=42).hexdigest()
        elif hash_size == 128:
            file_hash = xxhash.xxh128(data, seed=42).hexdigest()
        else:
            raise ValueError(f"Invalid hash size {hash_size}. Must be one of [32, 64, 128].")

    return file_hash


def compute_file_size_hash(path: str, hash_size: Literal[32, 64, 128] = 64) -> str:
    """Get hash of a file based on its size.

    Args:
        path: Path to the file.
        hash_size: Size of the hash. Must be one of [32, 64, 128].

    Returns:
        The hash of the file.
    """
    data = str(os.path.getsize(path))

    if hash_size == 32:
        file_hash = xxhash.xxh32(data, seed=42).hexdigest()
    elif hash_size == 64:
        file_hash = xxhash.xxh64(data, seed=42).hexdigest()
    elif hash_size == 128:
        file_hash = xxhash.xxh128(data, seed=42).hexdigest()
    else:
        raise ValueError(f"Invalid hash size {hash_size}. Must be one of [32, 64, 128].")

    return file_hash


@typing.no_type_check
def istarmap(self, func: Callable, iterable: Iterable, chunksize: int = 1):
    # pylint: disable=all
    """Starmap-version of imap."""
    self._check_running()
    if chunksize < 1:
        raise ValueError(f"Chunksize must be 1+, not {chunksize:n}")

    task_batches = mpp.Pool._get_tasks(func, iterable, chunksize)
    result = mpp.IMapIterator(self)
    self._taskqueue.put((self._guarded_task_generation(result._job, mpp.starmapstar, task_batches), result._set_length))
    return (item for chunk in result for item in chunk)


# Patch Pool class to include istarmap
mpp.Pool.istarmap = istarmap  # type: ignore[attr-defined]


class BaseDataModule(LightningDataModule, metaclass=DecorateParentMethod):
    """Base class for all data modules.

    Args:
        data_path: Path to the data main folder.
        name: The name for the data module. Defaults to "base_datamodule".
        num_workers: Number of workers for dataloaders. Defaults to 16.
        batch_size: Batch size. Defaults to 32.
        seed: Random generator seed. Defaults to 42.
        train_transform: Transformations for train dataset.
            Defaults to None.
        val_transform: Transformations for validation dataset.
            Defaults to None.
        test_transform: Transformations for test dataset.
            Defaults to None.
        enable_hashing: Whether to enable hashing of images. Defaults to True.
        hash_size: Size of the hash. Must be one of [32, 64, 128]. Defaults to 64.
        hash_type: Type of hash to use, if content hash is used, the hash is computed on the file content, otherwise
            the hash is computed on the file size which is faster but less safe. Defaults to "content".
    """

    def __init__(
        self,
        data_path: str,
        name: str = "base_datamodule",
        num_workers: int = 16,
        batch_size: int = 32,
        seed: int = 42,
        load_aug_images: bool = False,
        aug_name: str | None = None,
        n_aug_to_take: int | None = None,
        replace_str_from: str | None = None,
        replace_str_to: str | None = None,
        train_transform: albumentations.Compose | None = None,
        val_transform: albumentations.Compose | None = None,
        test_transform: albumentations.Compose | None = None,
        enable_hashing: bool = True,
        hash_size: Literal[32, 64, 128] = 64,
        hash_type: Literal["content", "size"] = "content",
    ):
        super().__init__()
        self.num_workers = num_workers
        self.batch_size = batch_size
        self.seed = seed
        self.data_path = data_path
        self.name = name
        self.train_transform = train_transform
        self.val_transform = val_transform
        self.test_transform = test_transform
        self.enable_hashing = enable_hashing
        self.hash_size = hash_size
        self.hash_type = hash_type

        if self.hash_size not in [32, 64, 128]:
            raise ValueError(f"Invalid hash size {self.hash_size}. Must be one of [32, 64, 128].")

        self.load_aug_images = load_aug_images
        self.aug_name = aug_name
        self.n_aug_to_take = n_aug_to_take
        self.replace_str_from = replace_str_from
        self.replace_str_to = replace_str_to
        self.extra_args: dict[str, Any] = {}
        self.train_dataset: TrainDataset
        self.val_dataset: ValDataset
        self.test_dataset: TestDataset
        self.data: pd.DataFrame
        self.data_folder = "data"
        os.makedirs(self.data_folder, exist_ok=True)
        self.datamodule_checkpoint_file = os.path.join(self.data_folder, "datamodule.pkl")
        self.dataset_file = os.path.join(self.data_folder, "dataset.csv")

    @property
    def train_data(self) -> pd.DataFrame:
        """Get train data."""
        if not hasattr(self, "data"):
            raise ValueError("`data` attribute is not set. Cannot load train data.")
        return self.data[self.data["split"] == "train"]

    @property
    def val_data(self) -> pd.DataFrame:
        """Get validation data."""
        if not hasattr(self, "data"):
            raise ValueError("`data` attribute is not set. Cannot load val data.")
        return self.data[self.data["split"] == "val"]

    @property
    def test_data(self) -> pd.DataFrame:
        """Get test data."""
        if not hasattr(self, "data"):
            raise ValueError("`data` attribute is not set. Cannot load test data.")
        return self.data[self.data["split"] == "test"]

    def _dataset_available(self, dataset_name: str) -> bool:
        """Checks if the dataset is available.

        Args:
            dataset_name : Name of the dataset attribute.

        Returns:
            True if the dataset is available, False otherwise.
        """
        available = hasattr(self, dataset_name) and getattr(self, dataset_name) is not None
        if available:
            dataset_attr = getattr(self, dataset_name)
            if isinstance(dataset_attr, list):
                available = all(len(d) > 0 for d in dataset_attr)
            else:
                available = len(dataset_attr) > 0
        return available

    @property
    def train_dataset_available(self) -> bool:
        """Checks if the train dataset is available."""
        return self._dataset_available("train_dataset")

    @property
    def val_dataset_available(self) -> bool:
        """Checks if the validation dataset is available."""
        return self._dataset_available("val_dataset")

    @property
    def test_dataset_available(self) -> bool:
        """Checks if the test dataset is available."""
        return self._dataset_available("test_dataset")

    def _prepare_data(self) -> None:
        """Prepares the data, this should have exactly the same logic as the prepare_data method
        of a LightningModule.
        """
        raise NotImplementedError(
            "This method must be implemented, it should contain all the logic that normally is "
            "contained in the prepare_data method of a LightningModule."
        )

    def hash_data(self) -> None:
        """Computes the hash of the files inside the datasets."""
        if not self.enable_hashing:
            return

        # TODO: We need to find a way to annotate the columns of data.
        paths_and_hash_length = zip(self.data["samples"], [self.hash_size] * len(self.data), strict=False)

        with mp.Pool(min(8, mp.cpu_count() - 1)) as pool:
            self.data["hash"] = list(
                tqdm(
                    pool.istarmap(  # type: ignore[attr-defined]
                        compute_file_content_hash if self.hash_type == "content" else compute_file_size_hash,
                        paths_and_hash_length,
                    ),
                    total=len(self.data),
                    desc="Computing hashes",
                )
            )

        self.data["hash_type"] = self.hash_type

    def prepare_data(self) -> None:
        """Prepares the data, should be overridden by subclasses."""
        if hasattr(self, "data"):
            return

        self._prepare_data()
        self.hash_data()
        self.save_checkpoint()

    def __getstate__(self) -> dict[str, Any]:
        """This method is called when pickling the object.
        It's useful to remove attributes that shouldn't be pickled.
        """
        state = self.__dict__.copy()
        if "trainer" in state:
            # Lightning injects the trainer in the datamodule, we don't want to pickle it.
            del state["trainer"]

        return state

    def save_checkpoint(self) -> None:
        """Saves the datamodule to disk, utility function that is called from prepare_data. We are required to save
        datamodule to disk because we can't assign attributes to the datamodule in prepare_data when working with
        multiple gpus.
        """
        if not os.path.exists(self.datamodule_checkpoint_file) and not os.path.exists(self.dataset_file):
            with open(self.datamodule_checkpoint_file, "wb") as f:
                pkl.dump(self, f)

            self.data.to_csv(self.dataset_file, index=False)
            log.info("Datamodule checkpoint saved to disk.")

        if "targets" in self.data:
            if isinstance(self.data["targets"].iloc[0], np.ndarray):
                # If we find a numpy array target it's very likely one hot encoded,
                # in that case we just print the number of train/val/test samples
                grouping = ["split"]
            else:
                grouping = ["split", "targets"]
            log.info("Dataset Info:")
            split_order = {"train": 0, "val": 1, "test": 2}
            log.info(
                "\n%s",
                self.data.groupby(grouping)
                .size()
                .to_frame()
                .reset_index()
                .sort_values(by=["split"], key=lambda x: x.map(split_order))
                .rename(columns={0: "count"})
                .to_string(index=False),
            )

    def restore_checkpoint(self) -> None:
        """Loads the data from disk, utility function that should be called from setup."""
        if hasattr(self, "data"):
            return

        if not os.path.isfile(self.datamodule_checkpoint_file):
            raise ValueError(f"Dataset file {self.datamodule_checkpoint_file} does not exist.")

        with open(self.datamodule_checkpoint_file, "rb") as f:
            checkpoint_datamodule = pkl.load(f)
            for key, value in checkpoint_datamodule.__dict__.items():
                setattr(self, key, value)

    # TODO: Check if this function can be removed
    def load_augmented_samples(
        self,
        samples: list[str],
        targets: list[Any],
        replace_str_from: str | None = None,
        replace_str_to: str | None = None,
        shuffle: bool = False,
    ) -> tuple[list[str], list[str]]:
        """Loads augmented samples."""
        if self.n_aug_to_take is None:
            raise ValueError("`n_aug_to_take` is not set. Cannot load augmented samples.")
        aug_samples = []
        aug_labels = []
        for sample, label in zip(samples, targets, strict=False):
            aug_samples.append(sample)
            aug_labels.append(label)
            final_sample = sample
            if replace_str_from is not None and replace_str_to is not None:
                final_sample = final_sample.replace(replace_str_from, replace_str_to)
            base, ext = os.path.splitext(final_sample)
            for k in range(self.n_aug_to_take):
                aug_samples.append(base + "_" + str(k + 1) + ext)
                aug_labels.append(label)
        samples = aug_samples
        targets = aug_labels
        if shuffle:
            idexs = np.arange(len(aug_samples))
            np.random.shuffle(idexs)
            samples = np.array(samples)[idexs].tolist()
            targets = np.array(targets)[idexs].tolist()
        return samples, targets



from .anomaly import AnomalyDataModule
from .classification import (
    ClassificationDataModule,
    MultilabelClassificationDataModule,
    SklearnClassificationDataModule,
)
from .patch import PatchSklearnClassificationDataModule
from .segmentation import SegmentationDataModule, SegmentationMulticlassDataModule
from .ssl import SSLDataModule

__all__ = [
    "AnomalyDataModule",
    "ClassificationDataModule",
    "SklearnClassificationDataModule",
    "SegmentationDataModule",
    "SegmentationMulticlassDataModule",
    "PatchSklearnClassificationDataModule",
    "MultilabelClassificationDataModule",
    "SSLDataModule",
]



from __future__ import annotations

from torch import nn

from quadra.utils.models import L2Norm


class BaseNetworkBuilder(nn.Module):
    """Baseline Feature Extractor, with the possibility to map features to an hypersphere.
        If hypershperical is True the classifier is ignored.

    Args:
        features_extractor: Feature extractor as a toch.nn.Module.
        pre_classifier: Pre classifier as a torch.nn.Module. Defaults to nn.Identity() if None.
        classifier: Classifier as a torch.nn.Module. Defaults to nn.Identity() if None.
        freeze: Whether to freeze the feature extractor. Defaults to True.
        hyperspherical: Whether to map features to an hypersphere. Defaults to False.
        flatten_features: Whether to flatten the features before the pre_classifier. May be required if your model
            is outputting a feature map rather than a vector. Defaults to True.
    """

    def __init__(
        self,
        features_extractor: nn.Module,
        pre_classifier: nn.Module | None = None,
        classifier: nn.Module | None = None,
        freeze: bool = True,
        hyperspherical: bool = False,
        flatten_features: bool = True,
    ):
        super().__init__()
        if pre_classifier is None:
            pre_classifier = nn.Identity()

        if classifier is None:
            classifier = nn.Identity()

        self.features_extractor = features_extractor
        self.freeze = freeze
        self.hyperspherical = hyperspherical
        self.pre_classifier = pre_classifier
        self.classifier = classifier
        self.flatten: bool = False
        self._hyperspherical: bool = False
        self.l2: L2Norm | None = None
        self.flatten_features = flatten_features

        self.freeze = freeze
        self.hyperspherical = hyperspherical

        if self.freeze:
            for p in self.features_extractor.parameters():
                p.requires_grad = False

    @property
    def freeze(self) -> bool:
        """Whether to freeze the feature extractor."""
        return self._freeze

    @freeze.setter
    def freeze(self, value: bool) -> None:
        """Whether to freeze the feature extractor."""
        for p in self.features_extractor.parameters():
            p.requires_grad = not value

        self._freeze = value

    @property
    def hyperspherical(self) -> bool:
        """Whether to map the extracted features into an hypersphere."""
        return self._hyperspherical

    @hyperspherical.setter
    def hyperspherical(self, value: bool) -> None:
        """Whether to map the extracted features into an hypersphere."""
        self._hyperspherical = value
        self.l2 = L2Norm() if value else None

    def forward(self, x):
        x = self.features_extractor(x)

        if self.flatten_features:
            x = x.view(x.size(0), -1)

        x = self.pre_classifier(x)

        if self.hyperspherical:
            x = self.l2(x)

        x = self.classifier(x)

        return x



from __future__ import annotations

import os
import random
from pathlib import Path

import albumentations as alb
import cv2
import numpy as np
import pandas as pd
from pandas import DataFrame
from torch import Tensor
from torch.utils.data import Dataset

from quadra.utils.utils import IMAGE_EXTENSIONS


def create_validation_set_from_test_set(samples: DataFrame, seed: int = 0) -> DataFrame:
    """Craete Validation Set from Test Set.

    This function creates a validation set from test set by splitting both
    normal and abnormal samples to two.

    Args:
        samples: Dataframe containing dataset info such as filenames, splits etc.
        seed: Random seed to ensure reproducibility. Defaults to 0.
    """
    if seed > 0:
        random.seed(seed)

    # Split normal images.
    normal_test_image_indices = samples.index[(samples.split == "test") & (samples.targets == "good")].to_list()
    num_normal_valid_images = len(normal_test_image_indices) // 2

    indices_to_sample = random.sample(population=normal_test_image_indices, k=num_normal_valid_images)
    samples.loc[indices_to_sample, "split"] = "val"

    # Split abnormal images.
    abnormal_test_image_indices = samples.index[(samples.split == "test") & (samples.targets != "good")].to_list()
    num_abnormal_valid_images = len(abnormal_test_image_indices) // 2

    indices_to_sample = random.sample(population=abnormal_test_image_indices, k=num_abnormal_valid_images)
    samples.loc[indices_to_sample, "split"] = "val"

    return samples


def split_normal_images_in_train_set(samples: DataFrame, split_ratio: float = 0.1, seed: int = 0) -> DataFrame:
    """Split normal images in train set.

        This function splits the normal images in training set and assigns the
        values to the test set. This is particularly useful especially when the
        test set does not contain any normal images.

        This is important because when the test set doesn't have any normal images,
        AUC computation fails due to having single class.

    Args:
        samples: Dataframe containing dataset info such as filenames, splits etc.
        split_ratio: Train-Test normal image split ratio. Defaults to 0.1.
        seed: Random seed to ensure reproducibility. Defaults to 0.

    Returns:
        Output dataframe where the part of the training set is assigned to test set.
    """
    if seed > 0:
        random.seed(seed)

    normal_train_image_indices = samples.index[(samples.split == "train") & (samples.targets == "good")].to_list()
    num_normal_train_images = len(normal_train_image_indices)
    num_normal_valid_images = int(num_normal_train_images * split_ratio)

    indices_to_split_from_train_set = random.sample(population=normal_train_image_indices, k=num_normal_valid_images)
    samples.loc[indices_to_split_from_train_set, "split"] = "test"

    return samples


def make_anomaly_dataset(
    path: Path,
    split: str | None = None,
    split_ratio: float = 0.1,
    seed: int = 0,
    mask_suffix: str | None = None,
    create_test_set_if_empty: bool = True,
) -> DataFrame:
    """Create dataframe by parsing a folder following the MVTec data file structure.

    The files are expected to follow the structure:
        path/to/dataset/split/label/image_filename.xyz
        path/to/dataset/ground_truth/label/mask_filename.png

    Masks MUST be png images, no other format is allowed
    Split can be either train/val/test

    This function creates a dataframe to store the parsed information based on the following format:
    |---|---------------|-------|---------|--------------|-----------------------------------------------|-------------|
    |   | path          | split | targets | samples      | mask_path                                     | label_index |
    |---|---------------|-------|---------|--------------|-----------------------------------------------|-------------|
    | 0 | datasets/name |  test |  defect | filename.xyz | ground_truth/defect/filename{mask_suffix}.png | 1           |
    |---|---------------|-------|---------|--------------|-----------------------------------------------|-------------|

    Args:
        path: Path to dataset
        split: Dataset split (i.e., either train or test). Defaults to None.
        split_ratio: Ratio to split normal training images and add to the
            test set in case test set doesn't contain any normal images.
            Defaults to 0.1.
        seed: Random seed to ensure reproducibility when splitting. Defaults to 0.
        mask_suffix: String to append to the base filename to get the mask name, by default for MVTec dataset masks
            are saved as imagename_mask.png in this case the parameter shoul be filled with "_mask"
        create_test_set_if_empty: If True, create a test set if the test set is empty.


    Example:
        The following example shows how to get training samples from MVTec bottle category:

        >>> root = Path('./MVTec')
        >>> category = 'bottle'
        >>> path = root / category
        >>> path
        PosixPath('MVTec/bottle')

        >>> samples = make_anomaly_dataset(path, split='train', split_ratio=0.1, seed=0)
        >>> samples.head()
           path         split label image_path                           mask_path                   label_index
        0  MVTec/bottle train good MVTec/bottle/train/good/105.png MVTec/bottle/ground_truth/good/105_mask.png 0
        1  MVTec/bottle train good MVTec/bottle/train/good/017.png MVTec/bottle/ground_truth/good/017_mask.png 0
        2  MVTec/bottle train good MVTec/bottle/train/good/137.png MVTec/bottle/ground_truth/good/137_mask.png 0
        3  MVTec/bottle train good MVTec/bottle/train/good/152.png MVTec/bottle/ground_truth/good/152_mask.png 0
        4  MVTec/bottle train good MVTec/bottle/train/good/109.png MVTec/bottle/ground_truth/good/109_mask.png 0

    Returns:
        An output dataframe containing samples for the requested split (ie., train or test)
    """
    samples_list = [
        (str(path),) + filename.parts[-3:]
        for filename in path.glob("**/*")
        if filename.is_file()
        and os.path.splitext(filename)[-1].lower() in IMAGE_EXTENSIONS
        and ".ipynb_checkpoints" not in str(filename)
    ]

    if len(samples_list) == 0:
        raise RuntimeError(f"Found 0 images in {path}")

    samples_list.sort()

    data = pd.DataFrame(samples_list, columns=["path", "split", "targets", "samples"])
    data = data[data.split != "ground_truth"]

    # Create mask_path column, masks MUST have png extension
    data["mask_path"] = (
        data.path
        + "/ground_truth/"
        + data.targets
        + "/"
        + data.samples.apply(lambda x: os.path.splitext(os.path.basename(x))[0])
        + (f"{mask_suffix}.png" if mask_suffix is not None else ".png")
    )

    # Modify image_path column by converting to absolute path
    data["samples"] = data.path + "/" + data.split + "/" + data.targets + "/" + data.samples

    # Split the normal images in training set if test set doesn't
    # contain any normal images. This is needed because AUC score
    # cannot be computed based on 1-class
    if sum((data.split == "test") & (data.targets == "good")) == 0 and create_test_set_if_empty:
        data = split_normal_images_in_train_set(data, split_ratio, seed)

    # Good images don't have mask
    data.loc[(data.split == "test") & (data.targets == "good"), "mask_path"] = ""

    # Create label index for normal (0), anomalous (1) and unknown (-1) images.
    data.loc[data.targets == "good", "label_index"] = 0
    data.loc[~data.targets.isin(["good", "unknown"]), "label_index"] = 1
    data.loc[data.targets == "unknown", "label_index"] = -1
    data.label_index = data.label_index.astype(int)

    # Get the data frame for the split.
    if split is not None and split in ["train", "val", "test"]:
        data = data[data.split == split]
        data = data.reset_index(drop=True)

    return data


class AnomalyDataset(Dataset):
    """Anomaly Dataset.

    Args:
        transform: Albumentations compose.
        task: ``classification`` or ``segmentation``
        samples: Pandas dataframe containing samples following the same structure created by make_anomaly_dataset
        valid_area_mask: Optional path to the mask to use to filter out the valid area of the image. If None, the
            whole image is considered valid.
        crop_area: Optional tuple of 4 integers (x1, y1, x2, y2) to crop the image to the specified area. If None, the
            whole image is considered valid.
    """

    def __init__(
        self,
        transform: alb.Compose,
        samples: DataFrame,
        task: str = "segmentation",
        valid_area_mask: str | None = None,
        crop_area: tuple[int, int, int, int] | None = None,
    ) -> None:
        self.task = task
        self.transform = transform

        self.samples = samples
        self.samples = self.samples.reset_index(drop=True)
        self.split = self.samples.split.unique()[0]

        self.crop_area = crop_area
        self.valid_area_mask: np.ndarray | None = None

        if valid_area_mask is not None:
            if not os.path.exists(valid_area_mask):
                raise RuntimeError(f"Valid area mask {valid_area_mask} does not exist.")

            self.valid_area_mask = cv2.imread(valid_area_mask, 0) > 0

    def __len__(self) -> int:
        """Get length of the dataset."""
        return len(self.samples)

    def __getitem__(self, index: int) -> dict[str, str | Tensor]:
        """Get dataset item for the index ``index``.

        Args:
            index: Index to get the item.

        Returns:
            Dict of image tensor during training.
            Otherwise, Dict containing image path, target path, image tensor, label and transformed bounding box.
        """
        item: dict[str, str | Tensor] = {}

        image_path = self.samples.samples.iloc[index]
        image = cv2.imread(image_path)
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

        original_image_shape = image.shape
        if self.valid_area_mask is not None:
            image = image * self.valid_area_mask[:, :, np.newaxis]

        if self.crop_area is not None:
            image = image[self.crop_area[1] : self.crop_area[3], self.crop_area[0] : self.crop_area[2]]

        label_index = self.samples.label_index[index]

        if self.split == "train":
            pre_processed = self.transform(image=image)
            item = {"image": pre_processed["image"], "label": label_index}
        elif self.split in ["val", "test"]:
            item["image_path"] = image_path
            item["label"] = label_index

            if self.task == "segmentation":
                mask_path = self.samples.mask_path[index]

                # If good images have no associated mask create an empty one
                if label_index == 0:
                    mask = np.zeros(shape=original_image_shape[:2])
                elif os.path.isfile(mask_path):
                    mask = cv2.imread(mask_path, flags=0) / 255.0
                else:
                    # We need ones in the mask to compute correctly at least image level f1 score
                    mask = np.ones(shape=original_image_shape[:2])

                if self.valid_area_mask is not None:
                    mask = mask * self.valid_area_mask

                if self.crop_area is not None:
                    mask = mask[self.crop_area[1] : self.crop_area[3], self.crop_area[0] : self.crop_area[2]]

                pre_processed = self.transform(image=image, mask=mask)

                item["mask_path"] = mask_path
                item["mask"] = pre_processed["mask"]
            else:
                pre_processed = self.transform(image=image)

            item["image"] = pre_processed["image"]
        return item



"""References:
- https://arxiv.org/pdf/1708.03888.pdf
- https://github.com/pytorch/pytorch/blob/1.6/torch/optim/sgd.py.
"""

from __future__ import annotations

from collections.abc import Callable

import torch
from torch.nn import Parameter
from torch.optim.optimizer import Optimizer, _RequiredParameter, required


class LARS(Optimizer):
    r"""Extends SGD in PyTorch with LARS scaling from the paper
    `Large batch training of Convolutional Networks <https://arxiv.org/pdf/1708.03888.pdf>`_.

    Args:
        params: iterable of parameters to optimize or dicts defining
            parameter groups
        lr: learning rate
        momentum: momentum factor (default: 0)
        weight_decay: weight decay (L2 penalty) (default: 0)
        dampening: dampening for momentum (default: 0)
        nesterov: enables Nesterov momentum (default: False)
        trust_coefficient: trust coefficient for computing LR (default: 0.001)
        eps: eps for division denominator (default: 1e-8).

    Example:
        >>> model = torch.nn.Linear(10, 1)
        >>> input = torch.Tensor(10)
        >>> target = torch.Tensor([1.])
        >>> loss_fn = lambda input, target: (input - target) ** 2
        >>> #
        >>> optimizer = LARS(model.parameters(), lr=0.1, momentum=0.9)
        >>> optimizer.zero_grad()
        >>> loss_fn(model(input), target).backward()
        >>> optimizer.step()

    .. note::
        The application of momentum in the SGD part is modified according to
        the PyTorch standards. LARS scaling fits into the equation in the
        following fashion.

        .. math::
            \begin{aligned}
                g_{t+1} & = \text{lars_lr} * (\beta * p_{t} + g_{t+1}), \\
                v_{t+1} & = \\mu * v_{t} + g_{t+1}, \\
                p_{t+1} & = p_{t} - \text{lr} * v_{t+1},
            \\end{aligned}

        where :math:`p`, :math:`g`, :math:`v`, :math:`\\mu` and :math:`\beta` denote the
        parameters, gradient, velocity, momentum, and weight decay respectively.
        The :math:`lars_lr` is defined by Eq. 6 in the paper.
        The Nesterov version is analogously modified.

    .. warning::
        Parameters with weight decay set to 0 will automatically be excluded from
        layer-wise LR scaling. This is to ensure consistency with papers like SimCLR
        and BYOL.
    """

    def __init__(
        self,
        params: list[Parameter],
        lr: _RequiredParameter = required,
        momentum: float = 0,
        dampening: float = 0,
        weight_decay: float = 0,
        nesterov: bool = False,
        trust_coefficient: float = 0.001,
        eps: float = 1e-8,
    ):
        if lr is not required and lr < 0.0:  # type: ignore[operator]
            raise ValueError(f"Invalid learning rate: {lr}")
        if momentum < 0.0:
            raise ValueError(f"Invalid momentum value: {momentum}")
        if weight_decay < 0.0:
            raise ValueError(f"Invalid weight_decay value: {weight_decay}")

        defaults = {
            "lr": lr,
            "momentum": momentum,
            "dampening": dampening,
            "weight_decay": weight_decay,
            "nesterov": nesterov,
        }
        if nesterov and (momentum <= 0 or dampening != 0):
            raise ValueError("Nesterov momentum requires a momentum and zero dampening")

        self.eps = eps
        self.trust_coefficient = trust_coefficient

        super().__init__(params, defaults)

    def __setstate__(self, state):
        super().__setstate__(state)

        for group in self.param_groups:
            group.setdefault("nesterov", False)

    @torch.no_grad()
    def step(self, closure: Callable | None = None):
        """Performs a single optimization step.

        Args:
            closure: A closure that reevaluates the model and returns the loss. Defaults to None.
        """
        loss = None
        if closure is not None:
            with torch.enable_grad():
                loss = closure()

        # exclude scaling for params with 0 weight decay
        for group in self.param_groups:
            weight_decay = group["weight_decay"]
            momentum = group["momentum"]
            dampening = group["dampening"]
            nesterov = group["nesterov"]

            for p in group["params"]:
                if p.grad is None:
                    continue

                d_p = p.grad
                p_norm = torch.norm(p.data)
                g_norm = torch.norm(p.grad.data)

                # lars scaling + weight decay part
                if weight_decay != 0 and p_norm != 0 and g_norm != 0:
                    lars_lr = p_norm / (g_norm + p_norm * weight_decay + self.eps)
                    lars_lr *= self.trust_coefficient

                    d_p = d_p.add(p, alpha=weight_decay)
                    d_p *= lars_lr

                # sgd part
                if momentum != 0:
                    param_state = self.state[p]
                    if "momentum_buffer" not in param_state:
                        buf = param_state["momentum_buffer"] = torch.clone(d_p).detach()
                    else:
                        buf = param_state["momentum_buffer"]
                        buf.mul_(momentum).add_(d_p, alpha=1 - dampening)
                    if nesterov:
                        d_p = d_p.add(buf, alpha=momentum)
                    else:
                        d_p = buf

                p.add_(d_p, alpha=-group["lr"])

        return loss



from __future__ import annotations

from typing import cast

import numpy as np
import torch
from scipy.optimize import linear_sum_assignment
from skimage.measure import label, regionprops  # pylint: disable=no-name-in-module

from quadra.utils.evaluation import dice


def _pad_to_shape(a: np.ndarray, shape: tuple, constant_values: int = 0) -> np.ndarray:
    """Pad lower - right with 0s
    Args:
        a: numpy array to pad
        shape: shape of the resulting np.array
        constant_values: value to pad.

    Returns:
        Padded array
    """
    y_, x_ = shape
    y, x = a.shape
    y_pad = y_ - y
    x_pad = x_ - x
    return np.pad(
        a,
        (
            (0, y_pad),
            (0, x_pad),
        ),
        mode="constant",
        constant_values=constant_values,
    )


def _get_iou(bboxes1: np.ndarray, bboxes2: np.ndarray, approx_iou: bool = False) -> np.ndarray:
    """Intersect over union
    Args:
        bboxes1: extracted bounding boxes
        bboxes2: ground truth
        approx_iou: flag to approximate.

    Returns:
        Intersect over union array
    """
    x11, y11, x12, y12 = np.split(bboxes1, 4, axis=1)
    x21, y21, x22, y22 = np.split(bboxes2, 4, axis=1)

    # determine the (x, y)-coordinates of the intersection rectangle
    xA = np.maximum(x11, x21.T)
    yA = np.maximum(y11, y21.T)
    xB = np.minimum(x12, x22.T)
    yB = np.minimum(y12, y22.T)

    # compute the area of intersection rectangle
    inter_area = np.maximum((xB - xA), 0) * np.maximum((yB - yA), 0)

    # compute the area of both the prediction and ground-truth rectangles
    box_a_area = (x12 - x11) * (y12 - y11)
    box_b_area = (x22 - x21) * (y22 - y21)

    if approx_iou:
        iou = inter_area / box_b_area.T
    else:
        iou = inter_area / (box_a_area + box_b_area.T - inter_area)

    return iou


def _get_dice_matrix(
    labels_pred: np.ndarray,
    n_labels_pred: int,
    labels_gt: np.ndarray,
    n_labels_gt: int,
) -> np.ndarray:
    """Create dice matrix
    Args:
        labels_pred: predicted label
        n_labels_pred: number of label predicted
        labels_gt: ground truth labels
        n_labels_gt: number of gt labels.

    Returns:
        Dice matrix
    """
    m = np.zeros((n_labels_pred, n_labels_gt))
    for i in range(n_labels_pred):
        pred = labels_pred == i + 1
        for j in range(n_labels_gt):
            gt = labels_gt == j + 1
            m[i, j] = dice(
                torch.Tensor(pred).unsqueeze(0).unsqueeze(0),
                torch.Tensor(gt).unsqueeze(0).unsqueeze(0),
                reduction="none",
            )
    return m


def segmentation_props(
    pred: np.ndarray, mask: np.ndarray
) -> tuple[float, float, float, float, list[float], float, int, int, int, int]:
    """Return some information regarding a segmentation task.

    Args:
        pred (np.ndarray[bool]): Prediction of a segmentation model as
            a binary image.
        mask (np.ndarray[bool]): Ground truth mask as binary image

    Returns:
        1-Dice(pred, mask) Given a matrix (a_ij) = (1-Dice)(prediction_i, ground_truth_j),
            where prediction_i is the i-th prediction connected component and
            ground_truth_j is the j-th ground truth connected component,
            I compute the LSA (Linear Sum Assignment) to find the optimal 1-to-1 assignment
            between predictions and ground truths that minimize the (1-Dice) score.
            Then, for every unique pair of (predictioni, ground_truthj) we compute Average
             (1-Dice)(predictioni, ground_truthj)
        Average (1-Dice)(predictioni, ground_truthj) between True Positives
            (that is predictions associated to a ground truth), which is gratis from a[i,j],
            where the average is computed w.r.t. the total number of True Positives found.
        Average IoU(predictioni, ground_truthj) between True Positives
            (that is predictions associated to a ground truth),
            where the average is computed w.r.t. the total number of True Positives found.
            The IoU is computed between the minimum enclosing bounding box
            of a prediction and a ground truth.
        Average area of False Positives
        Histogram of false positives
        Average area of False Negatives
        Number of True Positives (predictions associated to a ground truth)
        Number of False Positives (predictions without a ground truth associated)
                and their avg. area (avg is taken w.r.t. the total number of False Positives found)
        Number of False Negatives (ground truth without a predictions associated)
            and their avg. area (avg is taken w.r.t. the total number of False Negatives found)
        Number of labels in the mask.
    """
    labels_pred, n_labels_pred = label(pred, connectivity=2, return_num=True, background=0)
    labels_mask, n_labels_mask = label(mask, connectivity=2, return_num=True, background=0)

    labels_pred = cast(np.ndarray, labels_pred)
    labels_mask = cast(np.ndarray, labels_mask)
    n_labels_pred = cast(int, n_labels_pred)
    n_labels_mask = cast(int, n_labels_mask)

    props_pred = regionprops(labels_pred)
    props_mask = regionprops(labels_mask)
    pred_bbox = np.array([props_pred[i].bbox for i in range(len(props_pred))])
    mask_bbox = np.array([props_mask[i].bbox for i in range(len(props_mask))])

    global_dice = float(
        dice(
            torch.Tensor(pred).unsqueeze(0).unsqueeze(0),
            torch.Tensor(mask).unsqueeze(0).unsqueeze(0),
        ).item()
    )
    lsa_iou = 0.0
    lsa_dice = 0.0
    tp_num = 0
    fp_num = 0
    fn_num = 0
    fp_area = 0.0
    fn_area = 0.0
    fp_hist: list[float] = []
    if n_labels_pred > 0 and n_labels_mask > 0:
        dice_mat = _get_dice_matrix(labels_pred, n_labels_pred, labels_mask, n_labels_mask)
        # Thresholding over Dice scores
        dice_mat = np.where(dice_mat <= 0.9, dice_mat, 1.0)
        iou_mat = _get_iou(pred_bbox, mask_bbox, approx_iou=False)
        dice_mat_shape = dice_mat.shape
        max_dim = np.max(dice_mat_shape)
        # Add dummy Dices so LSA is unique and i can compute FP and FN
        dice_mat = _pad_to_shape(dice_mat, (max_dim, max_dim), 1)
        lsa = linear_sum_assignment(dice_mat, maximize=False)
        for row, col in zip(lsa[0], lsa[1], strict=False):
            # More preds than GTs --> False Positive
            if row < n_labels_pred and col >= n_labels_mask:
                min_row = pred_bbox[row][0]
                min_col = pred_bbox[row][1]
                h = pred_bbox[row][2] - min_row
                w = pred_bbox[row][3] - min_col
                fp_num += 1
                area = pred[min_row : min_row + h, min_col : min_col + w].sum()
                fp_area += area
                fp_hist.append(area)
                continue

            # More GTs than preds --> False Negative
            if col < n_labels_mask and row >= n_labels_pred:
                min_row = mask_bbox[col][0]
                min_col = mask_bbox[col][1]
                h = mask_bbox[col][2] - min_row
                w = mask_bbox[col][3] - min_col
                fn_num += 1
                fn_area += mask[min_row : min_row + h, min_col : min_col + w].sum()
                continue

            # Real True Positive: a prediction has been assigned to a gt
            # with at least a 1-Dice score of 0.9
            if dice_mat[row, col] <= 0.9:
                tp_num += 1
                lsa_iou += iou_mat[row, col]
                lsa_dice += dice_mat[row, col]
            else:
                # Here we have both a FP and a FN
                min_row = pred_bbox[row][0]
                min_col = pred_bbox[row][1]
                h = pred_bbox[row][2] - min_row
                w = pred_bbox[row][3] - min_col
                fp_num += 1
                area = pred[min_row : min_row + h, min_col : min_col + w].sum()
                fp_area += area
                fp_hist.append(area)

                min_row = mask_bbox[col][0]
                min_col = mask_bbox[col][1]
                h = mask_bbox[col][2] - min_row
                w = mask_bbox[col][3] - min_col
                fn_num += 1
                fn_area += mask[min_row : min_row + h, min_col : min_col + w].sum()
    elif len(pred_bbox) > 0 and len(mask_bbox) == 0:  # No GTs --> FP
        for p_bbox in pred_bbox:
            min_row = p_bbox[0]
            min_col = p_bbox[1]
            h = p_bbox[2] - min_row
            w = p_bbox[3] - min_col
            fp_num += 1
            # print("FP area:", pred[min_row : min_row + h, min_col : min_col + w].sum())
            area = pred[min_row : min_row + h, min_col : min_col + w].sum()
            fp_area += area
            fp_hist.append(area)
    elif len(pred_bbox) == 0 and len(mask_bbox) > 0:  # No preds --> FN
        for m_bbox in mask_bbox:
            min_row = m_bbox[0]
            min_col = m_bbox[1]
            h = m_bbox[2] - min_row
            w = m_bbox[3] - min_col
            fn_num += 1
            # print("FN area:", mask[min_row : min_row + h, min_col : min_col + w].sum())
            fn_area += mask[min_row : min_row + h, min_col : min_col + w].sum()
    return (
        global_dice,
        lsa_dice,
        lsa_iou,
        fp_area,
        fp_hist,
        fn_area,
        tp_num,
        fp_num,
        fn_num,
        n_labels_mask,
    )



from __future__ import annotations

import math
import warnings
from collections.abc import Callable
from typing import cast

import numpy as np
import timm
import torch
import torch.nn.functional as F
import tqdm
from pytorch_grad_cam import GradCAM
from scipy import ndimage
from sklearn.linear_model._base import ClassifierMixin
from timm.models.layers import DropPath
from timm.models.vision_transformer import Mlp
from torch import nn

from quadra.models.evaluation import (
    BaseEvaluationModel,
    ONNXEvaluationModel,
    TorchEvaluationModel,
    TorchscriptEvaluationModel,
)
from quadra.utils import utils
from quadra.utils.vit_explainability import VitAttentionGradRollout

log = utils.get_logger(__name__)


def net_hat(input_size: int, output_size: int) -> torch.nn.Sequential:
    """Create a linear layer with input and output neurons.

    Args:
        input_size: Number of input neurons
        output_size: Number of output neurons.

    Returns:
        A sequential containing a single Linear layer taking input neurons and producing output neurons

    """
    return torch.nn.Sequential(torch.nn.Linear(input_size, output_size))


def create_net_hat(dims: list[int], act_fun: Callable = torch.nn.ReLU, dropout_p: float = 0) -> torch.nn.Sequential:
    """Create a sequence of linear layers with activation functions and dropout.

    Args:
        dims: Dimension of hidden layers and output
        act_fun: activation function to use between layers, default ReLU
        dropout_p: Dropout probability. Defaults to 0.

    Returns:
        Sequence of linear layers of dimension specified by the input, each linear layer is followed
            by an activation function and optionally a dropout layer with the input probability
    """
    components: list[nn.Module] = []
    for i, _ in enumerate(dims[:-2]):
        if dropout_p > 0:
            components.append(torch.nn.Dropout(dropout_p))
        components.append(net_hat(dims[i], dims[i + 1]))
        components.append(act_fun())
    components.append(net_hat(dims[-2], dims[-1]))
    components.append(L2Norm())
    return torch.nn.Sequential(*components)


class L2Norm(torch.nn.Module):
    """Compute L2 Norm."""

    def forward(self, x: torch.Tensor):
        return x / torch.norm(x, p=2, dim=1, keepdim=True)


def init_weights(m):
    """Basic weight initialization."""
    classname = m.__class__.__name__
    if classname.find("Conv2d") != -1 or classname.find("ConvTranspose2d") != -1:
        nn.init.kaiming_uniform_(m.weight)
        nn.init.zeros_(m.bias)
    elif classname.find("BatchNorm") != -1:
        nn.init.normal_(m.weight, 1.0, 0.02)
        nn.init.zeros_(m.bias)
    elif classname.find("Linear") != -1:
        nn.init.xavier_normal_(m.weight)
        m.bias.data.fill_(0)


def get_feature(
    feature_extractor: torch.nn.Module | BaseEvaluationModel,
    dl: torch.utils.data.DataLoader,
    iteration_over_training: int = 1,
    gradcam: bool = False,
    classifier: ClassifierMixin | None = None,
    input_shape: tuple[int, int, int] | None = None,
    limit_batches: int | None = None,
) -> tuple[np.ndarray, np.ndarray, np.ndarray | None]:
    """Given a dataloader and a PyTorch model, extract features with the model and return features and labels.

    Args:
        dl: PyTorch dataloader
        feature_extractor: Pretrained PyTorch backbone
        iteration_over_training: Extract feature iteration_over_training times for each image
            (best if used with augmentation)
        gradcam: Whether to compute gradcams. Notice that it will slow the function
        classifier: Scikit-learn classifier
        input_shape: [H,W,C], backbone input shape, needed by classifier's pytorch wrapper
        limit_batches: Limit the number of batches to be processed

    Returns:
        Tuple containing:
            features: Model features
            labels: input_labels
            grayscale_cams: Gradcam output maps, None if gradcam arg is False
    """
    if isinstance(feature_extractor, TorchEvaluationModel | TorchscriptEvaluationModel):
        # If we are working with torch based evaluation models we need to extract the model
        feature_extractor = feature_extractor.model
    elif isinstance(feature_extractor, ONNXEvaluationModel):
        gradcam = False

    feature_extractor.eval()

    # Setup gradcam
    if gradcam:
        if not hasattr(feature_extractor, "features_extractor"):
            gradcam = False
        elif isinstance(feature_extractor.features_extractor, timm.models.resnet.ResNet):
            target_layers = [feature_extractor.features_extractor.layer4[-1]]
            cam = GradCAM(
                model=feature_extractor,
                target_layers=target_layers,
            )
            for p in feature_extractor.features_extractor.layer4[-1].parameters():
                p.requires_grad = True
        elif is_vision_transformer(feature_extractor.features_extractor):
            grad_rollout = VitAttentionGradRollout(
                feature_extractor.features_extractor,
                classifier=classifier,
                example_input=None if input_shape is None else torch.randn(1, *input_shape),
            )
        else:
            gradcam = False

        if not gradcam:
            log.warning("Gradcam not implemented for this backbone, it will not be computed")

    # Extract features from data

    for iteration in range(iteration_over_training):
        for i, b in enumerate(tqdm.tqdm(dl)):
            x1, y1 = b

            if hasattr(feature_extractor, "parameters"):
                # Move input to the correct device and dtype
                parameter = next(feature_extractor.parameters())
                x1 = x1.to(parameter.device).to(parameter.dtype)
            elif isinstance(feature_extractor, BaseEvaluationModel):
                x1 = x1.to(feature_extractor.device).to(feature_extractor.model_dtype)

            if gradcam:
                y_hat = cast(list[torch.Tensor] | tuple[torch.Tensor] | torch.Tensor, feature_extractor(x1).detach())
                # mypy can't detect that gradcam is true only if we have a features_extractor
                if is_vision_transformer(feature_extractor.features_extractor):  # type: ignore[union-attr]
                    grayscale_cam_low_res = grad_rollout(
                        input_tensor=x1, targets_list=y1
                    )  # TODO: We are using labels (y1) but it would be better to use preds
                    orig_shape = grayscale_cam_low_res.shape
                    new_shape = (orig_shape[0], x1.shape[2], x1.shape[3])
                    zoom_factors = tuple(np.array(new_shape) / np.array(orig_shape))
                    grayscale_cam = ndimage.zoom(grayscale_cam_low_res, zoom_factors, order=1)
                else:
                    grayscale_cam = cam(input_tensor=x1, targets=None)
                feature_extractor.zero_grad(set_to_none=True)  # type: ignore[union-attr]
            else:
                with torch.no_grad():
                    y_hat = cast(list[torch.Tensor] | tuple[torch.Tensor] | torch.Tensor, feature_extractor(x1))
                grayscale_cams = None

            if isinstance(y_hat, list | tuple):
                y_hat = y_hat[0].cpu()
            else:
                y_hat = y_hat.cpu()

            if torch.cuda.is_available():
                torch.cuda.empty_cache()

            if i == 0 and iteration == 0:
                features = torch.cat([y_hat], dim=0)
                labels = np.concatenate([y1])
                if gradcam:
                    grayscale_cams = grayscale_cam
            else:
                features = torch.cat([features, y_hat], dim=0)
                labels = np.concatenate([labels, y1], axis=0)
                if gradcam:
                    grayscale_cams = np.concatenate([grayscale_cams, grayscale_cam], axis=0)

            if limit_batches is not None and (i + 1) >= limit_batches:
                break

    return features.detach().numpy(), labels, grayscale_cams


def is_vision_transformer(model: torch.nn.Module) -> bool:
    """Verify if pytorch module is a Vision Transformer.
    This check is primarily needed for gradcam computation in classification tasks.

    Args:
        model: Model
    """
    return type(model).__name__ == "VisionTransformer"


def _no_grad_trunc_normal_(tensor: torch.Tensor, mean: float, std: float, a: float, b: float):
    """Cut & paste from PyTorch official master until it's in a few official releases - RW
    Method based on https://people.sc.fsu.edu/~jburkardt/presentations/truncated_normal.pdf.

    Args:
        tensor: an n-dimensional `torch.Tensor`
        mean: the mean of the normal distribution
        std: the standard deviation of the normal distribution
        a: the minimum cutoff
        b: the maximum cutoff
    """

    def norm_cdf(x: float):
        """Computes standard normal cumulative distribution function."""
        return (1.0 + math.erf(x / math.sqrt(2.0))) / 2.0

    if (mean < a - 2 * std) or (mean > b + 2 * std):
        warnings.warn(
            (
                "mean is more than 2 std from [a, b] in nn.init.trunc_normal_. "
                "The distribution of values may be incorrect."
            ),
            stacklevel=2,
        )

    with torch.no_grad():
        # Values are generated by using a truncated uniform distribution and
        # then using the inverse CDF for the normal distribution.
        # Get upper and lower cdf values
        l = norm_cdf((a - mean) / std)
        u = norm_cdf((b - mean) / std)

        # Uniformly fill tensor with values from [l, u], then translate to
        # [2l-1, 2u-1].
        tensor.uniform_(2 * l - 1, 2 * u - 1)

        # Use inverse cdf transform for normal distribution to get truncated
        # standard normal
        tensor.erfinv_()

        # Transform to proper mean, std
        tensor.mul_(std * math.sqrt(2.0))
        tensor.add_(mean)

        # Clamp to ensure it's in the proper range
        tensor.clamp_(min=a, max=b)
        return tensor


def trunc_normal_(tensor: torch.Tensor, mean: float = 0.0, std: float = 1.0, a: float = -2.0, b: float = 2.0):
    """Call `_no_grad_trunc_normal_` with `torch.no_grad()`.

    Args:
        tensor: an n-dimensional `torch.Tensor`
        mean: the mean of the normal distribution
        std: the standard deviation of the normal distribution
        a: the minimum cutoff
        b: the maximum cutoff
    """
    return _no_grad_trunc_normal_(tensor, mean, std, a, b)


def clip_gradients(model: nn.Module, clip: float) -> list[float]:
    """Args:
        model: The model
        clip: The clip value.

    Returns:
        The norms of the gradients
    """
    norms = []
    for _, p in model.named_parameters():
        if p.grad is not None:
            param_norm = p.grad.data.norm(2)
            norms.append(param_norm.item())
            clip_coef = clip / (param_norm + 1e-6)
            if clip_coef < 1:
                p.grad.data.mul_(clip_coef)
    return norms


# TODO: do not use this implementation for new models


class AttentionExtractor(torch.nn.Module):
    """General attention extractor.

    Args:
        model: Backbone model which contains the attention layer.
            attention_layer_name: Attention layer for extracting attention maps.
            Defaults to "attn_drop".
        attention_layer_name: Attention layer for extracting attention maps.
    """

    def __init__(self, model: torch.nn.Module, attention_layer_name: str = "attn_drop"):
        super().__init__()
        self.model = model
        modules = [module for module_name, module in self.model.named_modules() if attention_layer_name in module_name]
        if modules:
            modules[-1].register_forward_hook(self.get_attention)
        self.attentions = torch.zeros((1, 0))

    def clear(self):
        """Clear the grabbed attentions."""
        self.attentions = torch.zeros((1, 0))

    def get_attention(self, module: nn.Module, input_tensor: torch.Tensor, output: torch.Tensor):  # pylint: disable=unused-argument
        """Method to be registered to grab attentions."""
        self.attentions = output.detach().clone().cpu()

    @staticmethod
    def process_attention_maps(attentions: torch.Tensor, img_width: int, img_height: int) -> torch.Tensor:
        """Preprocess attentions maps to be visualized.

        Args:
            attentions: grabbed attentions
            img_width: image width
            img_height: image height

        Returns:
            torch.Tensor: preprocessed attentions, with the shape equal to the one of the image from
            which attentions has been computed
        """
        if len(attentions.shape) == 4:
            # vit
            # batch, heads, N, N (class atention layer)
            attentions = attentions[:, :, 0, 1:]  # batch, heads, height-1

        else:
            # xcit
            # batch, heads, N
            attentions = attentions[:, :, 1:]  # batch, heads, dim-1
        nh = attentions.shape[1]
        patch_size = int(math.sqrt(img_width * img_height / attentions.shape[-1]))
        w_featmap = img_width // patch_size
        h_featmap = img_height // patch_size

        # we keep only the output patch attention we dont want cls
        attentions = attentions.reshape(attentions.shape[0], nh, w_featmap, h_featmap)
        attentions = F.interpolate(attentions, scale_factor=patch_size, mode="nearest")
        return attentions

    def forward(self, t: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
        self.clear()
        out = self.model(t)
        return (out, self.attentions)  # torch.jit.trace does not complain


# TODO: do not use this implementation for new models


class PositionalEncoding1D(torch.nn.Module):
    """Standard sine-cosine positional encoding from https://arxiv.org/abs/2010.11929.

    Args:
        d_model: Embedding dimension
        temperature: Temperature for the positional encoding. Defaults to 10000.0.
        dropout: Dropout rate. Defaults to 0.0.
        max_len: Maximum length of the sequence. Defaults to 5000.
    """

    def __init__(self, d_model: int, temperature: float = 10000.0, dropout: float = 0.0, max_len: int = 5000):
        super().__init__()
        self.dropout: torch.nn.Dropout | torch.nn.Identity
        if dropout > 0:
            self.dropout = torch.nn.Dropout(p=dropout)
        else:
            self.dropout = torch.nn.Identity()

        position = torch.arange(max_len).unsqueeze(1)
        div_term = torch.exp(torch.arange(0, d_model, 2) * (-math.log(temperature) / d_model))
        self.pe = torch.zeros(max_len, 1, d_model)
        self.pe[:, 0, 0::2] = torch.sin(position * div_term)
        self.pe[:, 0, 1::2] = torch.cos(position * div_term)
        self.pe = self.pe.permute(1, 0, 2)
        self.pe = torch.nn.Parameter(self.pe)
        self.pe.requires_grad = False

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of the positional encoding.

        Args:
            x: torch tensor [batch_size, seq_len, embedding_dim].
        """
        x = x + self.pe[:, : x.size(1), :]
        return self.dropout(x)


class LSABlock(torch.nn.Module):
    """Local Self Attention Block from https://arxiv.org/abs/2112.13492.

    Args:
        dim: embedding dimension
        num_heads: number of attention heads
        mlp_ratio: ratio of mlp hidden dim to embedding dim
        qkv_bias: enable bias for qkv if True
        drop: dropout rate
        attn_drop: attention dropout rate
        drop_path: stochastic depth rate
        act_layer: activation layer
        norm_layer:: normalization layer
        mask_diagonal: whether to mask Q^T x K diagonal with -infinity so not to
            count self relationship between tokens. Defaults to True
        learnable_temperature: whether to use a learnable temperature as specified in
            https://arxiv.org/abs/2112.13492. Defaults to True.
    """

    def __init__(
        self,
        dim: int,
        num_heads: int,
        mlp_ratio: float = 4.0,
        qkv_bias: bool = False,
        drop: float = 0.0,
        attn_drop: float = 0.0,
        drop_path: float = 0.0,
        act_layer: type[nn.Module] = torch.nn.GELU,
        norm_layer: type[torch.nn.LayerNorm] = torch.nn.LayerNorm,
        mask_diagonal: bool = True,
        learnable_temperature: bool = True,
    ):
        super().__init__()
        self.norm1 = norm_layer(dim)
        self.attn = LocalSelfAttention(
            dim,
            num_heads=num_heads,
            qkv_bias=qkv_bias,
            attn_drop=attn_drop,
            proj_drop=drop,
            mask_diagonal=mask_diagonal,
            learnable_temperature=learnable_temperature,
        )
        # NOTE: drop path for stochastic depth, we shall see if this is better than dropout here
        self.drop_path = DropPath(drop_path) if drop_path > 0.0 else torch.nn.Identity()
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = Mlp(in_features=dim, hidden_features=mlp_hidden_dim, act_layer=act_layer, drop=drop)

    def forward(self, x):
        x = x + self.drop_path(self.attn(self.norm1(x)))
        x = x + self.drop_path(self.mlp(self.norm2(x)))
        return x


class LocalSelfAttention(torch.nn.Module):
    """Local Self Attention from https://arxiv.org/abs/2112.13492.

    Args:
        dim: embedding dimension.
        num_heads: number of attention heads.
        qkv_bias: enable bias for qkv if True.
        attn_drop: attention dropout rate.
        proj_drop: projection dropout rate.
        mask_diagonal: whether to mask Q^T x K diagonal with -infinity
            so not to count self relationship between tokens. Defaults to True.
        learnable_temperature: whether to use a learnable temperature as specified in
            https://arxiv.org/abs/2112.13492. Defaults to True.
    """

    def __init__(
        self,
        dim: int,
        num_heads: int = 8,
        qkv_bias: bool = False,
        attn_drop: float = 0.0,
        proj_drop: float = 0.0,
        mask_diagonal: bool = True,
        learnable_temperature: bool = True,
    ):
        super().__init__()
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.mask_diagonal = mask_diagonal
        if learnable_temperature:
            self.register_parameter("scale", torch.nn.Parameter(torch.tensor(head_dim**-0.5, requires_grad=True)))
        else:
            self.scale = head_dim**-0.5

        self.qkv = torch.nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = torch.nn.Dropout(attn_drop)
        self.proj = torch.nn.Linear(dim, dim)
        self.proj_drop = torch.nn.Dropout(proj_drop)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Computes the local self attention.

        Args:
            x: input tensor

        Returns:
            Output of the local self attention.
        """
        B, N, C = x.shape
        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
        q, k, v = qkv.unbind(0)  # make torchscript happy (cannot use tensor as tuple)

        attn = (q @ k.transpose(-2, -1)) * self.scale
        if self.mask_diagonal:
            attn[torch.eye(N, device=attn.device, dtype=torch.bool).repeat(B, self.num_heads, 1, 1)] = -float("inf")
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)
        return x
